from __future__ import annotations

from random import uniform

import pytest

from scitbx import matrix


def first_derivative(func, x, h):
    return (-func(x + 2 * h) + 8 * func(x + h) - 8 * func(x - h) + func(x - 2 * h)) / (
        12 * h
    )


def second_derivative(func, x, y=None, h=None):
    if y is None:
        A = func(x + 2 * h)
        B = func(x + h)
        C = func(x)
        D = func(x - h)
        E = func(x - 2 * h)
        return (-(1 / 12) * (A + E) + (4 / 3) * (B + D) - (5 / 2) * C) / h**2
    else:
        A = func(x - h, y - h)
        B = func(x - h, y)
        C = func(x, y - h)
        D = func(x, y)
        E = func(x, y + h)
        F = func(x + h, y)
        G = func(x + h, y + h)
        return (A - B - C + 2 * D - E - F + G) / (2 * h**2)


def generate_data():

    from random import seed

    seed(0)

    s2 = matrix.col(
        (uniform(0, 1), uniform(0, 1), uniform(0, 1))
    ).normalize() * uniform(0.9, 1.1)

    s0 = matrix.col((0, 0, 1))

    b1, b2, b3, b4, b5, b6 = (
        uniform(0.5, 3.5),
        uniform(0.0, 0.5),
        uniform(0.5, 3.5),
        uniform(0.0, 0.5),
        uniform(0.0, 0.5),
        uniform(0.5, 3.5),
    )

    M = matrix.sqr((b1, 0, 0, b2, b3, 0, b4, b5, b6))

    sigma = M * M.transpose()

    # ctot = randint(100, 1000)

    return sigma, s0, s2


def generate_testdata():
    for i in range(10):
        sigma, s0, (b1, b2, b3) = generate_data()
        yield sigma, s0, b1, b2, b3


def ds2_db(b1, b2, b3):
    return [matrix.col((1, 0, 0)), matrix.col((0, 1, 0)), matrix.col((0, 0, 1))]


def compute_s2(b1, b2, b3):
    return matrix.col((b1, b2, b3))


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_epsilon(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_epsilon(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        return s0.length() - s2.length()

    def f1(x):
        return compute_epsilon(x, b2, b3)

    def f2(x):
        return compute_epsilon(b1, x, b3)

    def f3(x):
        return compute_epsilon(b1, b2, x)

    h = 0.001

    depdb1_num = first_derivative(f1, b1, h)
    depdb2_num = first_derivative(f2, b2, h)
    depdb3_num = first_derivative(f3, b3, h)

    def compute_dep(s2, ds2):
        return -(1 / s2.length()) * s2.dot(ds2)

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    depdb1_cal = compute_dep(s2, ds2[0])
    depdb2_cal = compute_dep(s2, ds2[1])
    depdb3_cal = compute_dep(s2, ds2[2])

    assert abs(depdb1_num - depdb1_cal) < 1e-7
    assert abs(depdb2_num - depdb2_cal) < 1e-7
    assert abs(depdb3_num - depdb3_cal) < 1e-7


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_mubar(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_mubar(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        epsilon = s0.length() - s2.length()
        return S12 * (1 / S22) * epsilon

    def f1(x):
        return compute_mubar(x, b2, b3)

    def f2(x):
        return compute_mubar(b1, x, b3)

    def f3(x):
        return compute_mubar(b1, b2, x)

    h = 0.001

    dmubar_db1_num = first_derivative(f1, b1, h)
    dmubar_db2_num = first_derivative(f2, b2, h)
    dmubar_db3_num = first_derivative(f3, b3, h)

    def compute_depsilon(s2, ds2):
        return -(1 / s2.length()) * s2.dot(ds2)

    def compute_dmubar(s2, ds2):
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        return S12 * (1 / S22) * compute_depsilon(s2, ds2)

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    dmubar_db1_cal = compute_dmubar(s2, ds2[0])
    dmubar_db2_cal = compute_dmubar(s2, ds2[1])
    dmubar_db3_cal = compute_dmubar(s2, ds2[2])

    assert abs(dmubar_db1_num - dmubar_db1_cal) < 1e-7
    assert abs(dmubar_db2_num - dmubar_db2_cal) < 1e-7
    assert abs(dmubar_db3_num - dmubar_db3_cal) < 1e-7


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_e1(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_e1(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        return s2.cross(s0).normalize()

    def f1(x):
        return compute_e1(x, b2, b3)

    def f2(x):
        return compute_e1(b1, x, b3)

    def f3(x):
        return compute_e1(b1, b2, x)

    h = 0.001

    de1db1_num = first_derivative(f1, b1, h)
    de1db2_num = first_derivative(f2, b2, h)
    de1db3_num = first_derivative(f3, b3, h)

    def compute_de1(s2, ds2):
        e1 = s2.cross(s0)
        d1 = ds2.cross(s0)
        A = d1 / e1.length()
        B = e1 * d1.dot(e1) / (e1.length() ** 3)
        return A - B

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    de1db1_cal = compute_de1(s2, ds2[0])
    de1db2_cal = compute_de1(s2, ds2[1])
    de1db3_cal = compute_de1(s2, ds2[2])

    assert abs(de1db1_num - de1db1_cal) < 1e-7
    assert abs(de1db2_num - de1db2_cal) < 1e-7
    assert abs(de1db3_num - de1db3_cal) < 1e-7


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_e2(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_e2(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        e1 = s2.cross(s0).normalize()
        return s2.cross(e1).normalize()

    def f1(x):
        return compute_e2(x, b2, b3)

    def f2(x):
        return compute_e2(b1, x, b3)

    def f3(x):
        return compute_e2(b1, b2, x)

    h = 0.001

    de2db1_num = first_derivative(f1, b1, h)
    de2db2_num = first_derivative(f2, b2, h)
    de2db3_num = first_derivative(f3, b3, h)

    def compute_de2(s2, ds2):
        e1 = s2.cross(s0)
        e2 = s2.cross(e1)
        d2 = ds2 * s2.dot(s0) + s2 * ds2.dot(s0) - 2 * s0 * s2.dot(ds2)
        A = d2 / e2.length()
        B = e2 * d2.dot(e2) / (e2.length() ** 3)
        return A - B

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    de2db1_cal = compute_de2(s2, ds2[0])
    de2db2_cal = compute_de2(s2, ds2[1])
    de2db3_cal = compute_de2(s2, ds2[2])

    assert abs(de2db1_num - de2db1_cal) < 1e-7
    assert abs(de2db2_num - de2db2_cal) < 1e-7
    assert abs(de2db3_num - de2db3_cal) < 1e-7


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_e3(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_e3(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        return s2.normalize()

    def f1(x):
        return compute_e3(x, b2, b3)

    def f2(x):
        return compute_e3(b1, x, b3)

    def f3(x):
        return compute_e3(b1, b2, x)

    h = 0.001

    de3db1_num = first_derivative(f1, b1, h)
    de3db2_num = first_derivative(f2, b2, h)
    de3db3_num = first_derivative(f3, b3, h)

    def compute_de3(s2, ds2):
        e3 = s2
        d3 = ds2
        A = d3 / e3.length()
        B = e3 * d3.dot(e3) / (e3.length() ** 3)
        return A - B

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    de3db1_cal = compute_de3(s2, ds2[0])
    de3db2_cal = compute_de3(s2, ds2[1])
    de3db3_cal = compute_de3(s2, ds2[2])

    assert abs(de3db1_num - de3db1_cal) < 1e-7
    assert abs(de3db2_num - de3db2_cal) < 1e-7
    assert abs(de3db3_num - de3db3_cal) < 1e-7


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_s1(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_s1(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        epsilon = s0.length() - s2.length()
        mubar = S12 * (1 / S22) * epsilon

        e1 = s2.cross(s0).normalize()
        e2 = s2.cross(e1).normalize()
        e3 = s2.normalize()
        R = matrix.sqr(e1.elems + e2.elems + e3.elems)
        s1 = R.transpose() * matrix.col((mubar[0], mubar[1], s0.length()))
        return s1

    def f1(x):
        return compute_s1(x, b2, b3)

    def f2(x):
        return compute_s1(b1, x, b3)

    def f3(x):
        return compute_s1(b1, b2, x)

    h = 0.001

    db1_num = first_derivative(f1, b1, h)
    db2_num = first_derivative(f2, b2, h)
    db3_num = first_derivative(f3, b3, h)

    def compute_depsilon(s2, ds2):
        return -(1 / s2.length()) * s2.dot(ds2)

    def compute_dmubar(s2, ds2):
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        return S12 * (1 / S22) * compute_depsilon(s2, ds2)

    def compute_R(s2):
        e1 = s2.cross(s0).normalize()
        e2 = s2.cross(e1).normalize()
        e3 = s2.normalize()
        R = matrix.sqr(e1.elems + e2.elems + e3.elems)
        return R

    def compute_de1(s2, ds2):
        e1 = s2.cross(s0)
        d1 = ds2.cross(s0)
        A = d1 / e1.length()
        B = e1 * d1.dot(e1) / (e1.length() ** 3)
        return A - B

    def compute_de2(s2, ds2):
        e1 = s2.cross(s0)
        e2 = s2.cross(e1)
        d2 = ds2 * s2.dot(s0) + s2 * ds2.dot(s0) - 2 * s0 * s2.dot(ds2)
        A = d2 / e2.length()
        B = e2 * d2.dot(e2) / (e2.length() ** 3)
        return A - B

    def compute_de3(s2, ds2):
        e3 = s2
        d3 = ds2
        A = d3 / e3.length()
        B = e3 * d3.dot(e3) / (e3.length() ** 3)
        return A - B

    def compute_dR(s2, ds2):
        de1 = compute_de1(s2, ds2)
        de2 = compute_de2(s2, ds2)
        de3 = compute_de3(s2, ds2)
        return matrix.sqr(de1.elems + de2.elems + de3.elems)

    def compute_ds1(s2, ds2):
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        R = compute_R(s2)
        dR = compute_dR(s2, ds2)
        epsilon = s0.length() - s2.length()
        mubar = S12 * (1 / S22) * epsilon
        dmubar = compute_dmubar(s2, ds2)
        v = matrix.col((mubar[0], mubar[1], s0.length()))
        A = -R.transpose() * dR * R.transpose() * v
        B = R.transpose() * matrix.col((dmubar[0], dmubar[1], 0))
        return A + B

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    db1_cal = compute_ds1(s2, ds2[0])
    db2_cal = compute_ds1(s2, ds2[1])
    db3_cal = compute_ds1(s2, ds2[2])

    assert abs(db1_num - db1_cal) < 1e-7
    assert abs(db2_num - db2_cal) < 1e-7
    assert abs(db3_num - db3_cal) < 1e-7


@pytest.mark.parametrize("sigma,s0,b1,b2,b3", generate_testdata())
def test_derivative_of_f(sigma, s0, b1, b2, b3):

    ds2 = ds2_db(b1, b2, b3)

    def compute_f(b1, b2, b3):
        s2 = compute_s2(b1, b2, b3)
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        epsilon = s0.length() - s2.length()
        mubar = S12 * (1 / S22) * epsilon

        e1 = s2.cross(s0).normalize()
        e2 = s2.cross(e1).normalize()
        e3 = s2.normalize()
        R = matrix.sqr(e1.elems + e2.elems + e3.elems)
        s1 = R.transpose() * matrix.col((mubar[0], mubar[1], s0.length()))

        D = matrix.sqr((1, 0, 0, 0, 1, 0, 0, 0, 1))

        v = D * s1
        assert v[2] > 0
        X = v[0] / v[2]
        Y = v[1] / v[2]

        xcal = matrix.col((X, Y))
        xobs = matrix.col((0, 0))
        return (xobs - xcal).transpose() * (xobs - xcal)

    def f1(x):
        return compute_f(x, b2, b3)

    def f2(x):
        return compute_f(b1, x, b3)

    def f3(x):
        return compute_f(b1, b2, x)

    h = 0.001

    db1_num = first_derivative(f1, b1, h)
    db2_num = first_derivative(f2, b2, h)
    db3_num = first_derivative(f3, b3, h)

    def compute_depsilon(s2, ds2):
        return -(1 / s2.length()) * s2.dot(ds2)

    def compute_dmubar(s2, ds2):
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        return S12 * (1 / S22) * compute_depsilon(s2, ds2)

    def compute_R(s2):
        e1 = s2.cross(s0).normalize()
        e2 = s2.cross(e1).normalize()
        e3 = s2.normalize()
        R = matrix.sqr(e1.elems + e2.elems + e3.elems)
        return R

    def compute_de1(s2, ds2):
        e1 = s2.cross(s0)
        d1 = ds2.cross(s0)
        A = d1 / e1.length()
        B = e1 * d1.dot(e1) / (e1.length() ** 3)
        return A - B

    def compute_de2(s2, ds2):
        e1 = s2.cross(s0)
        e2 = s2.cross(e1)
        d2 = ds2 * s2.dot(s0) + s2 * ds2.dot(s0) - 2 * s0 * s2.dot(ds2)
        A = d2 / e2.length()
        B = e2 * d2.dot(e2) / (e2.length() ** 3)
        return A - B

    def compute_de3(s2, ds2):
        e3 = s2
        d3 = ds2
        A = d3 / e3.length()
        B = e3 * d3.dot(e3) / (e3.length() ** 3)
        return A - B

    def compute_dR(s2, ds2):
        de1 = compute_de1(s2, ds2)
        de2 = compute_de2(s2, ds2)
        de3 = compute_de3(s2, ds2)
        return matrix.sqr(de1.elems + de2.elems + de3.elems)

    def compute_ds1(s2, ds2):
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        R = compute_R(s2)
        dR = compute_dR(s2, ds2)
        epsilon = s0.length() - s2.length()
        mubar = S12 * (1 / S22) * epsilon
        dmubar = compute_dmubar(s2, ds2)
        v = matrix.col((mubar[0], mubar[1], s0.length()))
        A = -R.transpose() * dR * R.transpose() * v
        B = R.transpose() * matrix.col((dmubar[0], dmubar[1], 0))
        return A + B

    def compute_s1(s2):
        S12 = matrix.col((sigma[2], sigma[5]))
        S22 = sigma[8]
        epsilon = s0.length() - s2.length()
        mubar = S12 * (1 / S22) * epsilon

        e1 = s2.cross(s0).normalize()
        e2 = s2.cross(e1).normalize()
        e3 = s2.normalize()
        R = matrix.sqr(e1.elems + e2.elems + e3.elems)
        s1 = R.transpose() * matrix.col((mubar[0], mubar[1], s0.length()))
        return s1

    def compute_df(s2, ds2):
        s1 = compute_s1(s2)
        ds1 = compute_ds1(s2, ds2)
        D = matrix.sqr((1, 0, 0, 0, 1, 0, 0, 0, 1))
        v = D * s1
        # dv = D * ds1
        assert v[2] > 0
        X = v[0] / v[2]
        Y = v[1] / v[2]
        xcal = matrix.col((X, Y))
        dx = (v[2] * ds1[0] - v[0] * ds1[2]) / v[2] ** 2
        dy = (v[2] * ds1[1] - v[1] * ds1[2]) / v[2] ** 2
        xobs = matrix.col((0, 0))
        dxy = matrix.col((dx, dy))
        return -2 * (xobs - xcal).transpose() * dxy

    s2 = compute_s2(b1, b2, b3)
    ds2 = ds2_db(b1, b2, b3)

    db1_cal = compute_df(s2, ds2[0])
    db2_cal = compute_df(s2, ds2[1])
    db3_cal = compute_df(s2, ds2[2])

    assert abs(db1_num - db1_cal) < 1e-7
    assert abs(db2_num - db2_cal) < 1e-7
    assert abs(db3_num - db3_cal) < 1e-7



from __future__ import annotations

import copy
import logging
import math

import libtbx
from dxtbx.model.experiment_list import Experiment, ExperimentList

from dials.algorithms.indexing import DialsIndexError, DialsIndexRefineError
from dials.algorithms.indexing.indexer import Indexer
from dials.algorithms.indexing.known_orientation import IndexerKnownOrientation
from dials.algorithms.indexing.lattice_search import BasisVectorSearch, LatticeSearch
from dials.algorithms.indexing.nave_parameters import NaveParameters
from dials.array_family import flex
from dials.util.multi_dataset_handling import generate_experiment_identifiers

logger = logging.getLogger(__name__)


def calc_2D_rmsd_and_displacements(reflections):
    displacements = flex.vec2_double(
        reflections["xyzobs.px.value"].parts()[0],
        reflections["xyzobs.px.value"].parts()[1],
    ) - flex.vec2_double(
        reflections["xyzcal.px"].parts()[0], reflections["xyzcal.px"].parts()[1]
    )
    rmsd = math.sqrt(flex.mean(displacements.dot(displacements)))

    return rmsd, displacements


def plot_displacements(reflections, predictions, experiments):
    rmsd, displacements = calc_2D_rmsd_and_displacements(predictions)

    from matplotlib import pyplot as plt

    plt.figure()
    for cv in displacements:
        plt.plot([cv[0]], [-cv[1]], "r.")
    plt.title(f" {len(displacements)} spots, r.m.s.d. {rmsd:5.2f} pixels")
    plt.axes().set_aspect("equal")
    plt.show()
    plt.close()

    plt.figure()
    sz1, sz2 = experiments[0].detector[0].get_image_size()
    for item, cv in zip(predictions, displacements):
        plt.plot([item["xyzcal.px"][0]], [sz1 - item["xyzcal.px"][1]], "r.")
        plt.plot([item["xyzobs.px.value"][0]], [sz1 - item["xyzobs.px.value"][1]], "g.")
        plt.plot(
            [item["xyzcal.px"][0], item["xyzcal.px"][0] + 10.0 * cv[0]],
            [sz1 - item["xyzcal.px"][1], sz1 - item["xyzcal.px"][1] - 10.0 * cv[1]],
            "r-",
        )

    plt.xlim([0, experiments[0].detector[0].get_image_size()[0]])
    plt.ylim([0, experiments[0].detector[0].get_image_size()[1]])
    plt.title(f" {len(displacements)} spots, r.m.s.d. {rmsd:5.2f} pixels")
    plt.axes().set_aspect("equal")
    plt.show()
    plt.close()


def e_refine(params, experiments, reflections, graph_verbose=False):
    # Stills-specific parameters we always want
    assert params.refinement.reflections.outlier.algorithm in (
        None,
        "null",
    ), "Cannot index, set refinement.reflections.outlier.algorithm=null"  # we do our own outlier rejection

    from dials.algorithms.refinement.refiner import RefinerFactory

    refiner = RefinerFactory.from_parameters_data_experiments(
        params, reflections, experiments
    )

    refiner.run()

    ref_sel = refiner.selection_used_for_refinement()
    assert ref_sel.count(True) == len(reflections)

    if not graph_verbose:
        return refiner

    RR = refiner.predict_for_reflection_table(reflections)
    plot_displacements(reflections, RR, experiments)

    return refiner


class StillsIndexer(Indexer):
    """Class for indexing stills"""

    def __init__(self, reflections, experiments, params=None):
        if params.refinement.reflections.outlier.algorithm in ("auto", libtbx.Auto):
            # The stills_indexer provides its own outlier rejection
            params.refinement.reflections.outlier.algorithm = "null"
        super().__init__(reflections, experiments, params)

    def index(self):
        # most of this is the same as dials.algorithms.indexing.indexer.indexer_base.index(), with some stills
        # specific modifications (don't re-index after choose best orientation matrix, but use the indexing from
        # choose best orientation matrix, also don't use macrocycles) of refinement after indexing.
        # 2017 update: do accept multiple lattices per shot

        experiments = ExperimentList()

        while True:
            self.d_min = self.params.refinement_protocol.d_min_start
            max_lattices = self.params.multiple_lattice_search.max_lattices
            if max_lattices is not None and len(experiments) >= max_lattices:
                break
            if len(experiments) > 0:
                cutoff_fraction = (
                    self.params.multiple_lattice_search.recycle_unindexed_reflections_cutoff
                )
                d_spacings = 1 / self.reflections["rlp"].norms()
                d_min_indexed = flex.min(d_spacings.select(self.indexed_reflections))
                min_reflections_for_indexing = cutoff_fraction * len(
                    self.reflections.select(d_spacings > d_min_indexed)
                )
                crystal_ids = self.reflections.select(d_spacings > d_min_indexed)["id"]
                if (crystal_ids == -1).count(True) < min_reflections_for_indexing:
                    logger.info(
                        "Finish searching for more lattices: %i unindexed reflections remaining.",
                        min_reflections_for_indexing,
                    )
                    break

            n_lattices_previous_cycle = len(experiments)

            # index multiple lattices per shot
            if len(experiments) == 0:
                new = self.find_lattices()
                generate_experiment_identifiers(new)
                experiments.extend(new)
                if len(experiments) == 0:
                    raise DialsIndexError("No suitable lattice could be found.")
            else:
                try:
                    new = self.find_lattices()
                    generate_experiment_identifiers(new)
                    experiments.extend(new)
                except Exception as e:
                    logger.info("Indexing remaining reflections failed")
                    logger.debug(
                        "Indexing remaining reflections failed, exception:\n" + str(e)
                    )

            # reset reflection lattice flags
            # the lattice a given reflection belongs to: a value of -1 indicates
            # that a reflection doesn't belong to any lattice so far
            self.reflections["id"] = flex.int(len(self.reflections), -1)

            self.index_reflections(experiments, self.reflections)

            if len(experiments) == n_lattices_previous_cycle:
                # no more lattices found
                break

            if (
                not self.params.stills.refine_candidates_with_known_symmetry
                and self.params.known_symmetry.space_group is not None
            ):
                self._apply_symmetry_post_indexing(
                    experiments, self.reflections, n_lattices_previous_cycle
                )

            # discard nearly overlapping lattices on the same shot
            if self._check_have_similar_crystal_models(experiments):
                break

            self.indexed_reflections = self.reflections["id"] > -1
            if self.d_min is None:
                sel = self.reflections["id"] <= -1
            else:
                sel = flex.bool(len(self.reflections), False)
                lengths = 1 / self.reflections["rlp"].norms()
                isel = (lengths >= self.d_min).iselection()
                sel.set_selected(isel, True)
                sel.set_selected(self.reflections["id"] > -1, False)
            self.unindexed_reflections = self.reflections.select(sel)

            reflections_for_refinement = self.reflections.select(
                self.indexed_reflections
            )

            if len(self.params.stills.isoforms) > 0:
                logger.info("")
                logger.info("#" * 80)
                logger.info("Starting refinement")
                logger.info("#" * 80)
                logger.info("")

                isoform_experiments = ExperimentList()
                isoform_reflections = flex.reflection_table()
                # Note, changes to params after initial indexing. Cannot use tie to target when fixing the unit cell.
                self.all_params.refinement.reflections.outlier.algorithm = "null"
                self.all_params.refinement.parameterisation.crystal.fix = "cell"
                self.all_params.refinement.parameterisation.crystal.unit_cell.restraints.tie_to_target = (
                    []
                )

                for expt_id, experiment in enumerate(experiments):
                    reflections = reflections_for_refinement.select(
                        reflections_for_refinement["id"] == expt_id
                    )
                    reflections["id"] = flex.int(len(reflections), 0)
                    refiners = []
                    for isoform in self.params.stills.isoforms:
                        iso_experiment = copy.deepcopy(experiment)
                        crystal = iso_experiment.crystal
                        if (
                            isoform.lookup_symbol
                            != crystal.get_space_group().type().lookup_symbol()
                        ):
                            logger.info(
                                "Crystal isoform lookup_symbol %s does not match isoform %s lookup_symbol %s",
                                crystal.get_space_group().type().lookup_symbol(),
                                isoform.name,
                                isoform.lookup_symbol,
                            )
                            continue
                        crystal.set_B(isoform.cell.fractionalization_matrix())

                        logger.info("Refining isoform %s", isoform.name)
                        refiners.append(
                            e_refine(
                                params=self.all_params,
                                experiments=ExperimentList([iso_experiment]),
                                reflections=reflections,
                                graph_verbose=False,
                            )
                        )

                    if len(refiners) == 0:
                        raise DialsIndexError(
                            "No isoforms had a lookup symbol that matched"
                        )
                    positional_rmsds = [
                        math.sqrt(P.rmsds()[0] ** 2 + P.rmsds()[1] ** 2)
                        for P in refiners
                    ]
                    logger.info(
                        "Positional rmsds for all isoforms:" + str(positional_rmsds)
                    )
                    minrmsd_mm = min(positional_rmsds)
                    minindex = positional_rmsds.index(minrmsd_mm)
                    logger.info(
                        "The smallest rmsd is %5.1f um from isoform %s",
                        1000.0 * minrmsd_mm,
                        self.params.stills.isoforms[minindex].name,
                    )
                    if self.params.stills.isoforms[minindex].rmsd_target_mm is not None:
                        logger.info(
                            "Asserting %f < %f",
                            minrmsd_mm,
                            self.params.stills.isoforms[minindex].rmsd_target_mm,
                        )
                        assert (
                            minrmsd_mm
                            < self.params.stills.isoforms[minindex].rmsd_target_mm
                        )
                    logger.info(
                        "Acceptable rmsd for isoform %s.",
                        self.params.stills.isoforms[minindex].name,
                    )
                    if len(self.params.stills.isoforms) == 2:
                        logger.info(
                            "Rmsd gain over the other isoform %5.1f um.",
                            1000.0 * abs(positional_rmsds[0] - positional_rmsds[1]),
                        )
                    R = refiners[minindex]
                    # Now one last check to see if direct beam is out of bounds
                    if self.params.stills.isoforms[minindex].beam_restraint is not None:
                        from scitbx import matrix

                        refined_beam = matrix.col(
                            R.get_experiments()[0]
                            .detector[0]
                            .get_beam_centre_lab(experiments[0].beam.get_s0())[0:2]
                        )
                        known_beam = matrix.col(
                            self.params.stills.isoforms[minindex].beam_restraint
                        )
                        logger.info(
                            "Asserting difference in refined beam center and expected beam center %f < %f",
                            (refined_beam - known_beam).length(),
                            self.params.stills.isoforms[minindex].rmsd_target_mm,
                        )
                        assert (
                            refined_beam - known_beam
                        ).length() < self.params.stills.isoforms[
                            minindex
                        ].rmsd_target_mm
                        # future--circle of confusion could be given as a separate length in mm instead of reusing rmsd_target

                    experiment = R.get_experiments()[0]
                    experiment.crystal.identified_isoform = self.params.stills.isoforms[
                        minindex
                    ].name

                    isoform_experiments.append(experiment)
                    reflections["id"] = flex.int(len(reflections), expt_id)
                    isoform_reflections.extend(reflections)
                experiments = isoform_experiments
                reflections_for_refinement = isoform_reflections

            if self.params.refinement_protocol.mode == "repredict_only":

                from dials.algorithms.indexing.nave_parameters import NaveParameters
                from dials.algorithms.refinement.prediction.managed_predictors import (
                    ExperimentsPredictorFactory,
                )

                refined_experiments, refined_reflections = (
                    experiments,
                    reflections_for_refinement,
                )
                ref_predictor = ExperimentsPredictorFactory.from_experiments(
                    experiments,
                    force_stills=True,
                    spherical_relp=self.all_params.refinement.parameterisation.spherical_relp_model,
                )
                ref_predictor(refined_reflections)
                refined_reflections["delpsical2"] = (
                    refined_reflections["delpsical.rad"] ** 2
                )
                for expt_id in range(len(refined_experiments)):
                    refls = refined_reflections.select(
                        refined_reflections["id"] == expt_id
                    )
                    nv = NaveParameters(
                        params=self.all_params,
                        experiments=refined_experiments[expt_id : expt_id + 1],
                        reflections=refls,
                        refinery=None,
                        graph_verbose=False,
                    )
                    experiments[expt_id].crystal = nv()
                ref_predictor = ExperimentsPredictorFactory.from_experiments(
                    experiments,
                    force_stills=True,
                    spherical_relp=self.all_params.refinement.parameterisation.spherical_relp_model,
                )
                ref_predictor(refined_reflections)

            elif self.params.refinement_protocol.mode is None:
                refined_experiments, refined_reflections = (
                    experiments,
                    reflections_for_refinement,
                )

            else:
                try:
                    refined_experiments, refined_reflections = self.refine(
                        experiments, reflections_for_refinement
                    )
                except Exception as e:
                    s = str(e)
                    if len(experiments) == 1:
                        raise DialsIndexRefineError(e)
                    logger.info("Refinement failed:")
                    logger.info(s)
                    del experiments[-1]
                    break

            self._unit_cell_volume_sanity_check(experiments, refined_experiments)

            self.refined_reflections = refined_reflections.select(
                refined_reflections["id"] > -1
            )

            for i, expt in enumerate(self.experiments):
                ref_sel = self.refined_reflections.select(
                    self.refined_reflections["imageset_id"] == i
                )
                ref_sel = ref_sel.select(ref_sel["id"] >= 0)
                for i_expt in set(ref_sel["id"]):
                    refined_expt = refined_experiments[i_expt]
                    expt.detector = refined_expt.detector
                    expt.beam = refined_expt.beam
                    expt.goniometer = refined_expt.goniometer
                    expt.scan = refined_expt.scan
                    refined_expt.imageset = expt.imageset

            if not (
                self.all_params.refinement.parameterisation.beam.fix == "all"
                and self.all_params.refinement.parameterisation.detector.fix == "all"
            ):
                # Experimental geometry may have changed - re-map centroids to
                # reciprocal space
                self.reflections.map_centroids_to_reciprocal_space(self.experiments)

            # update for next cycle
            experiments = refined_experiments
            self.refined_experiments = refined_experiments

        if self.refined_experiments is None:
            raise DialsIndexRefineError("None of the experiments could refine.")

        # discard experiments with zero reflections after refinement
        id_set = set(self.refined_reflections["id"])
        if len(id_set) < len(self.refined_experiments):
            filtered_refined_reflections = flex.reflection_table()
            for i in range(len(self.refined_experiments)):
                if i not in id_set:
                    del self.refined_experiments[i]
            for old, new in zip(sorted(id_set), range(len(id_set))):
                subset = self.refined_reflections.select(
                    self.refined_reflections["id"] == old
                )
                subset["id"] = flex.int(len(subset), new)
                filtered_refined_reflections.extend(subset)
            self.refined_reflections = filtered_refined_reflections

        if len(self.refined_experiments) > 1:
            from dials.algorithms.indexing.compare_orientation_matrices import (
                rotation_matrix_differences,
            )

            logger.info(
                rotation_matrix_differences(self.refined_experiments.crystals())
            )

        logger.info("Final refined crystal models:")
        for i, crystal_model in enumerate(self.refined_experiments.crystals()):
            n_indexed = 0
            for _ in experiments.where(crystal=crystal_model):
                n_indexed += (self.reflections["id"] == i).count(True)
            logger.info("model %i (%i reflections):", i + 1, n_indexed)
            logger.info(crystal_model)

        if (
            "xyzcal.mm" in self.refined_reflections
        ):  # won't be there if refine_all_candidates = False and no isoforms

            self._xyzcal_mm_to_px(self.experiments, self.refined_reflections)

    def experiment_list_for_crystal(self, crystal):
        experiments = ExperimentList()
        for imageset in self.experiments.imagesets():
            experiments.append(
                Experiment(
                    imageset=imageset,
                    beam=imageset.get_beam(),
                    detector=imageset.get_detector(),
                    goniometer=imageset.get_goniometer(),
                    scan=imageset.get_scan(),
                    crystal=crystal,
                )
            )
        return experiments

    def choose_best_orientation_matrix(self, candidate_orientation_matrices):
        logger.info("*" * 80)
        logger.info("Selecting the best orientation matrix")
        logger.info("*" * 80)

        class CandidateInfo(libtbx.group_args):
            pass

        candidates = []

        params = copy.deepcopy(self.all_params)

        for icm, cm in enumerate(candidate_orientation_matrices):
            if icm >= self.params.basis_vector_combinations.max_refine:
                break
            # Index reflections in P1
            sel = self.reflections["id"] == -1
            refl = self.reflections.select(sel)
            experiments = self.experiment_list_for_crystal(cm)
            self.index_reflections(experiments, refl)
            indexed = refl.select(refl["id"] >= 0)
            indexed = indexed.select(indexed.get_flags(indexed.flags.indexed))

            # If target symmetry supplied, try to apply it.  Then, apply the change of basis to the reflections
            # indexed in P1 to the target setting
            if (
                self.params.stills.refine_candidates_with_known_symmetry
                and self.params.known_symmetry.space_group is not None
            ):
                new_crystal, cb_op = self._symmetry_handler.apply_symmetry(cm)
                if new_crystal is None:
                    logger.info("Cannot convert to target symmetry, candidate %d", icm)
                    continue
                cm = new_crystal.change_basis(cb_op)
                experiments = self.experiment_list_for_crystal(cm)

                if not cb_op.is_identity_op():
                    indexed["miller_index"] = cb_op.apply(indexed["miller_index"])

            if params.indexing.stills.refine_all_candidates:
                try:
                    logger.info(
                        "$$$ stills_indexer::choose_best_orientation_matrix, candidate %d initial outlier identification",
                        icm,
                    )
                    acceptance_flags = self.identify_outliers(
                        params, experiments, indexed
                    )
                    # create a new "indexed" list with outliers thrown out:
                    indexed = indexed.select(acceptance_flags)

                    logger.info(
                        "$$$ stills_indexer::choose_best_orientation_matrix, candidate %d refinement before outlier rejection",
                        icm,
                    )
                    R = e_refine(
                        params=params,
                        experiments=experiments,
                        reflections=indexed,
                        graph_verbose=False,
                    )
                    ref_experiments = R.get_experiments()

                    # try to improve the outcome with a second round of outlier rejection post-initial refinement:
                    acceptance_flags = self.identify_outliers(
                        params, ref_experiments, indexed
                    )

                    # insert a round of Nave-outlier rejection on top of the r.m.s.d. rejection
                    nv0 = NaveParameters(
                        params=params,
                        experiments=ref_experiments,
                        reflections=indexed,
                        refinery=R,
                        graph_verbose=False,
                    )
                    nv0()
                    acceptance_flags_nv0 = nv0.nv_acceptance_flags
                    indexed = indexed.select(acceptance_flags & acceptance_flags_nv0)

                    logger.info(
                        "$$$ stills_indexer::choose_best_orientation_matrix, candidate %d after positional and delta-psi outlier rejection",
                        icm,
                    )
                    R = e_refine(
                        params=params,
                        experiments=ref_experiments,
                        reflections=indexed,
                        graph_verbose=False,
                    )
                    ref_experiments = R.get_experiments()

                    nv = NaveParameters(
                        params=params,
                        experiments=ref_experiments,
                        reflections=indexed,
                        refinery=R,
                        graph_verbose=False,
                    )
                    crystal_model = nv()
                    assert (
                        len(crystal_model) == 1
                    ), "$$$ stills_indexer::choose_best_orientation_matrix, Only one crystal at this stage"
                    crystal_model = crystal_model[0]

                    # Drop candidates that after refinement can no longer be converted to the known target space group
                    if (
                        not self.params.stills.refine_candidates_with_known_symmetry
                        and self.params.known_symmetry.space_group is not None
                    ):
                        (
                            new_crystal,
                            cb_op_to_primitive,
                        ) = self._symmetry_handler.apply_symmetry(crystal_model)
                        if new_crystal is None:
                            logger.info(
                                "P1 refinement yielded model diverged from target, candidate %d",
                                icm,
                            )
                            continue

                    rmsd, _ = calc_2D_rmsd_and_displacements(
                        R.predict_for_reflection_table(indexed)
                    )
                except Exception as e:
                    logger.info(
                        "Couldn't refine candidate %d, %s: %s",
                        icm,
                        e.__class__.__name__,
                        str(e),
                    )
                else:
                    logger.info(
                        "$$$ stills_indexer::choose_best_orientation_matrix, candidate %d done",
                        icm,
                    )
                    candidates.append(
                        CandidateInfo(
                            crystal=crystal_model,
                            green_curve_area=nv.green_curve_area,
                            ewald_proximal_volume=nv.ewald_proximal_volume(),
                            n_indexed=len(indexed),
                            rmsd=rmsd,
                            indexed=indexed,
                            experiments=ref_experiments,
                        )
                    )
            else:
                from dials.algorithms.refinement.prediction.managed_predictors import (
                    ExperimentsPredictorFactory,
                )

                ref_predictor = ExperimentsPredictorFactory.from_experiments(
                    experiments,
                    force_stills=True,
                    spherical_relp=params.refinement.parameterisation.spherical_relp_model,
                )
                rmsd, _ = calc_2D_rmsd_and_displacements(ref_predictor(indexed))
                candidates.append(
                    CandidateInfo(
                        crystal=cm,
                        n_indexed=len(indexed),
                        rmsd=rmsd,
                        indexed=indexed,
                        experiments=experiments,
                    )
                )
        if len(candidates) == 0:
            raise DialsIndexError("No suitable indexing solution found")

        logger.info("**** ALL CANDIDATES:")
        for i, XX in enumerate(candidates):
            logger.info("\n****Candidate %d %s", i, XX)
            cc = XX.crystal
            if hasattr(cc, "get_half_mosaicity_deg"):
                logger.info(
                    "  half mosaicity %5.2f deg.", (cc.get_half_mosaicity_deg())
                )
                logger.info("  domain size %.0f Ang.", (cc.get_domain_size_ang()))
        logger.info("\n**** BEST CANDIDATE:")

        results = flex.double([c.rmsd for c in candidates])
        best = candidates[flex.min_index(results)]
        logger.info(best)

        if params.indexing.stills.refine_all_candidates:
            if best.rmsd > params.indexing.stills.rmsd_min_px:
                raise DialsIndexError(f"RMSD too high, {best.rmsd:f}")

            if len(candidates) > 1:
                for i in range(len(candidates)):
                    if i == flex.min_index(results):
                        continue
                    if best.ewald_proximal_volume > candidates[i].ewald_proximal_volume:
                        logger.info(
                            "Couldn't figure out which candidate is best; picked the one with the best RMSD."
                        )

        best.indexed["entering"] = flex.bool(best.n_indexed, False)

        return best.crystal, best.n_indexed

    def identify_outliers(self, params, experiments, indexed):
        if not params.indexing.stills.candidate_outlier_rejection:
            return flex.bool(len(indexed), True)

        logger.info("$$$ stills_indexer::identify_outliers")
        refiner = e_refine(params, experiments, indexed, graph_verbose=False)

        RR = refiner.predict_for_reflection_table(indexed)

        px_sz = experiments[0].detector[0].get_pixel_size()

        class Match:
            pass

        matches = []
        for item in RR.rows():
            m = Match()
            m.x_obs = item["xyzobs.px.value"][0] * px_sz[0]
            m.y_obs = item["xyzobs.px.value"][1] * px_sz[1]
            m.x_calc = item["xyzcal.px"][0] * px_sz[0]
            m.y_calc = item["xyzcal.px"][1] * px_sz[1]
            m.miller_index = item["miller_index"]
            matches.append(m)

        import iotbx.phil
        from rstbx.phil.phil_preferences import indexing_api_defs

        hardcoded_phil = iotbx.phil.parse(input_string=indexing_api_defs).extract()

        from rstbx.indexing_api.outlier_procedure import OutlierPlotPDF

        # comment this in if PDF graph is desired:
        # hardcoded_phil.indexing.outlier_detection.pdf = "outlier.pdf"
        # new code for outlier rejection inline here
        if hardcoded_phil.indexing.outlier_detection.pdf is not None:
            hardcoded_phil.__inject__(
                "writer", OutlierPlotPDF(hardcoded_phil.indexing.outlier_detection.pdf)
            )

        # execute Sauter and Poon (2010) algorithm
        from rstbx.indexing_api import outlier_detection

        od = outlier_detection.find_outliers_from_matches(
            matches, verbose=True, horizon_phil=hardcoded_phil
        )

        if hardcoded_phil.indexing.outlier_detection.pdf is not None:
            od.make_graphs(canvas=hardcoded_phil.writer.R.c, left_margin=0.5)
            hardcoded_phil.writer.R.c.showPage()
            hardcoded_phil.writer.R.c.save()

        return od.get_cache_status()

    def refine(self, experiments, reflections):
        acceptance_flags = self.identify_outliers(
            self.all_params, experiments, reflections
        )
        # create a new "reflections" list with outliers thrown out:
        reflections = reflections.select(acceptance_flags)

        R = e_refine(
            params=self.all_params,
            experiments=experiments,
            reflections=reflections,
            graph_verbose=False,
        )
        ref_experiments = R.get_experiments()

        # try to improve the outcome with a second round of outlier rejection post-initial refinement:
        acceptance_flags = self.identify_outliers(
            self.all_params, ref_experiments, reflections
        )

        # insert a round of Nave-outlier rejection on top of the r.m.s.d. rejection
        nv0 = NaveParameters(
            params=self.all_params,
            experiments=ref_experiments,
            reflections=reflections,
            refinery=R,
            graph_verbose=False,
        )
        nv0()
        acceptance_flags_nv0 = nv0.nv_acceptance_flags
        reflections = reflections.select(acceptance_flags & acceptance_flags_nv0)

        R = e_refine(
            params=self.all_params,
            experiments=ref_experiments,
            reflections=reflections,
            graph_verbose=False,
        )
        ref_experiments = R.get_experiments()

        nv = NaveParameters(
            params=self.all_params,
            experiments=ref_experiments,
            reflections=reflections,
            refinery=R,
            graph_verbose=False,
        )
        nv()
        rmsd, _ = calc_2D_rmsd_and_displacements(
            R.predict_for_reflection_table(reflections)
        )

        matches = R.get_matches()
        xyzcal_mm = flex.vec3_double(len(reflections))
        xyzcal_mm.set_selected(matches["iobs"], matches["xyzcal.mm"])
        reflections["xyzcal.mm"] = xyzcal_mm
        reflections.set_flags(matches["iobs"], reflections.flags.used_in_refinement)
        reflections["entering"] = flex.bool(len(reflections), False)

        if self.all_params.indexing.stills.set_domain_size_ang_value is not None:
            for exp in ref_experiments:
                exp.crystal.set_domain_size_ang(
                    self.all_params.indexing.stills.set_domain_size_ang_value
                )
        if self.all_params.indexing.stills.set_mosaic_half_deg_value is not None:
            for exp in ref_experiments:
                exp.crystal.set_half_mosaicity_deg(
                    self.all_params.indexing.stills.set_mosaic_half_deg_value
                )

        return ref_experiments, reflections


"""Mixin class definitions that override the dials indexing class methods specific to stills"""


class StillsIndexerKnownOrientation(IndexerKnownOrientation, StillsIndexer):
    pass


class StillsIndexerBasisVectorSearch(StillsIndexer, BasisVectorSearch):
    pass


class StillsIndexerLatticeSearch(StillsIndexer, LatticeSearch):
    pass



from __future__ import annotations

import pytest

import scitbx.matrix
from cctbx import sgtbx
from cctbx.sgtbx import bravais_types
from dxtbx.model import Crystal, Experiment, ExperimentList

from dials.algorithms.indexing import assign_indices, non_primitive_basis
from dials.array_family import flex


@pytest.mark.parametrize("space_group_symbol", bravais_types.acentric)
def test_detect(space_group_symbol):
    sgi = sgtbx.space_group_info(space_group_symbol)
    cs = sgi.any_compatible_crystal_symmetry(volume=1000)
    ms = cs.build_miller_set(anomalous_flag=True, d_min=1).expand_to_p1()
    result = non_primitive_basis.detect(ms.indices())
    if sgi.group().conventional_centring_type_symbol() != "P":
        assert result is not None
        assert isinstance(result, scitbx.matrix.sqr)
        assert result.n == (3, 3)
    else:
        assert result is None


@pytest.mark.parametrize("space_group_symbol", bravais_types.acentric)
def test_correct(space_group_symbol):

    sgi = sgtbx.space_group_info(space_group_symbol)
    cs = sgi.any_compatible_crystal_symmetry(volume=1000)
    ms = cs.build_miller_set(anomalous_flag=True, d_min=1).expand_to_p1()

    # the reciprocal matrix
    B = scitbx.matrix.sqr(cs.unit_cell().fractionalization_matrix()).transpose()
    crystal = Crystal(B, sgtbx.space_group())
    expts = ExperimentList([Experiment(crystal=crystal)])

    refl = flex.reflection_table()
    refl["miller_index"] = ms.indices()
    refl["rlp"] = B.elems * ms.indices().as_vec3_double()
    refl["imageset_id"] = flex.int(len(refl))
    refl["xyzobs.mm.value"] = flex.vec3_double(len(refl))

    non_primitive_basis.correct(expts, refl, assign_indices.AssignIndicesGlobal())

    cs_corrected = expts.crystals()[0].get_crystal_symmetry()
    assert cs_corrected.change_of_basis_op_to_primitive_setting().is_identity_op()
    assert (
        cs.change_of_basis_op_to_primitive_setting().apply(ms.indices())
        == refl["miller_index"]
    )



from __future__ import annotations

import os

import procrunner


def test_goniometer_calibration(dials_regression, tmpdir):
    data_dir = os.path.join(dials_regression, "rotation_calibration")
    o0_k0_p0 = os.path.join(data_dir, "experiments_o0_k0_p0.json")
    o0_k0_p48 = os.path.join(data_dir, "experiments_o0_k0_p48.json")
    o0_k48_p48 = os.path.join(data_dir, "experiments_o0_k48_p48.json")
    o48_k48_p48 = os.path.join(data_dir, "experiments_o48_k48_p48.json")

    command = [
        "dials.goniometer_calibration",
        o0_k0_p0,
        o0_k0_p48,
        o0_k48_p48,
        o48_k48_p48,
        "space_group=P422",
        "xoalign=xoalign_config.py",
    ]

    print(command)
    result = procrunner.run(command, working_directory=tmpdir)
    assert not result.returncode and not result.stderr

    expected_output = b"""
Goniometer axes and angles (ImgCIF coordinate system):
GON_PHI:  rotation of 47.976 degrees about axis (0.99997,-0.00588,-0.00586)
GON_KAPPA:  rotation of 47.773 degrees about axis (0.91314,0.27943,-0.29681)
GON_OMEGA:  rotation of 48.000 degrees about axis (1.00000,0.00000,0.00000)

Goniometer axes and angles (MOSFLM coordinate system):
GON_PHI:  rotation of 47.976 degrees about axis (0.00494,-0.00588,0.99997)
GON_KAPPA:  rotation of 47.773 degrees about axis (0.29597,0.27943,0.91341)
GON_OMEGA:  rotation of 48.000 degrees about axis (-0.00092,0.00000,1.00000)

ImgCIF _axis loop template:
loop_
  _axis.id
  _axis.type
  _axis.equipment
  _axis.depends_on
  _axis.vector[1]
  _axis.vector[2]
  _axis.vector[3]
  _axis.offset[1]
  _axis.offset[2]
  _axis.offset[3]
  GON_OMEGA  rotation  goniometer  .          1.0000   0.0000   0.0000  .  .  .
  GON_KAPPA  rotation  goniometer  GON_OMEGA  0.9131   0.2794  -0.2968  .  .  .
  GON_PHI    rotation  goniometer  GON_KAPPA  1.0000  -0.0059  -0.0059  .  .  .

"""
    assert result.stdout.replace(b"\r\n", b"\n").endswith(expected_output)

    assert tmpdir.join("xoalign_config.py").check()
    expected_xoalign_config = """\
GONIOMETER_AXES_NAMES = ('GON_OMEGA', 'GON_KAPPA', 'GON_PHI')
GONIOMETER_AXES = [(-0.00092, 0.00000, 1.00000), (0.29597, 0.27943, 0.91341), (0.00494, -0.00588, 0.99997)]
GONIOMETER_DATUM = (0,0,0) # in degrees
"""
    xoalign_config = tmpdir.join("xoalign_config.py").read()
    assert xoalign_config == expected_xoalign_config



from __future__ import annotations

import libtbx.phil

import dials.util

help_message = """

Compute resolution-wise distribution of spots

Examples::

  dials.spot_resolution_shells models.expt strong.refl
"""

phil_scope = libtbx.phil.parse(
    """
  shells = 100
    .type = int(value_min=1)
"""
)


def settings():
    return phil_scope.fetch().extract()


def spot_resolution_shells(experiments, reflections, params):
    from dials.array_family import flex

    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    two_theta_array = reflections["rlp"].norms()
    h0 = flex.weighted_histogram(flex.pow2(two_theta_array), n_slots=params.shells)
    n = h0.slots()
    d = 1.0 / flex.sqrt(h0.slot_centers())

    for j in range(params.shells):
        print("%d %f %d" % (j, d[j], n[j]))


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )

    usage = "dials.spot_resolution_shells [options] models.expt observations.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 or len(reflections) == 0:
        parser.print_help()
        exit(0)

    reflections = reflections[0]

    spot_resolution_shells(experiments, reflections, params)


if __name__ == "__main__":
    run()



"""
Tests for scaling utilities module.
"""

from __future__ import annotations

from math import pi, sqrt

import numpy as np
import pytest

from dxtbx.model import (
    Beam,
    Crystal,
    Experiment,
    ExperimentList,
    GoniometerFactory,
    Scan,
)
from dxtbx.serialize import load
from libtbx import phil
from scitbx.sparse import matrix  # noqa: F401 - Needed to call calc_theta_phi

from dials.algorithms.scaling.scaling_library import create_scaling_model
from dials.algorithms.scaling.scaling_utilities import (
    Reasons,
    align_axis_along_z,
    calc_crystal_frame_vectors,
    calculate_prescaling_correction,
    quasi_normalisation,
    set_wilson_outliers,
)
from dials.array_family import flex
from dials.util.options import ArgumentParser
from dials_scaling_ext import (
    calc_lookup_index,
    calc_theta_phi,
    calculate_harmonic_tables_from_selections,
    create_sph_harm_lookup_table,
    create_sph_harm_table,
)


@pytest.fixture(scope="module")
def test_experiment_singleaxisgonio():
    gonio = GoniometerFactory.from_dict(
        {
            "axes": [
                [1.0, 0.0, 0.0],
            ],
            "angles": [0.0],
            "names": ["GON_PHI"],
            "scan_axis": 0,
        }
    )
    return Experiment(
        beam=Beam(s0=(0.0, 0.0, 2.0)),
        goniometer=gonio,
        scan=Scan(image_range=[1, 90], oscillation=[0.0, 1.0]),
    )


def test_experiments_multiaxisgonio():
    """Create a mock experiments object."""
    # beam along +z
    gonio_1 = GoniometerFactory.from_dict(
        {
            "axes": [
                [
                    1.0 / sqrt(2.0),
                    0.0,
                    -1.0 / sqrt(2.0),
                ],
                [1.0, 0.0, 0.0],
            ],
            "angles": [0.0, 0.0],
            "names": ["GON_PHI", "GON_OMEGA"],
            "scan_axis": 1,
        }
    )
    gonio_2 = GoniometerFactory.from_dict(
        {
            "axes": [
                [
                    1.0 / sqrt(2.0),
                    0.0,
                    -1.0 / sqrt(2.0),
                ],
                [1.0, 0.0, 0.0],
            ],
            "angles": [0.0, 0.0],
            "names": ["GON_PHI", "GON_OMEGA"],
            "scan_axis": 0,
        }
    )

    experiments = ExperimentList()
    for g in [gonio_1, gonio_2]:
        experiments.append(
            Experiment(
                beam=Beam(s0=(0.0, 0.0, 2.0)),
                goniometer=g,
                scan=Scan(image_range=[1, 90], oscillation=[0.0, 1.0]),
            )
        )

    return experiments


@pytest.fixture(scope="module")
def test_exp_E2():
    """Create a mock experiments object."""
    exp = Experiment()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " P 1",
    }
    crystal = Crystal.from_dict(exp_dict)
    exp.crystal = crystal
    return exp


@pytest.fixture(scope="module")
def test_exp_P1():
    """Create a mock experiments object."""
    exp = Experiment()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 1.0],
        "space_group_hall_symbol": " P 1",
    }
    crystal = Crystal.from_dict(exp_dict)
    exp.crystal = crystal
    return exp


@pytest.fixture(scope="module")
def test_reflection_table():
    """Return a test reflection table."""
    return generate_reflection_table()


@pytest.fixture(scope="module")
def wilson_test_reflection_table():
    """Return a test reflection table."""
    rt = flex.reflection_table()
    rt["centric_flag"] = flex.bool([True, True, False, False])
    rt["Esq"] = flex.double([50.0, 10.0, 50.0, 10.0])
    return rt


def generate_reflection_table():
    """Create a reflection table with s1 and phi."""
    rt = flex.reflection_table()
    s1_vec = (1.0 / sqrt(2.0), 0.0, 1.0 / sqrt(2.0))
    rt["s1"] = flex.vec3_double([s1_vec, s1_vec, s1_vec])
    rt["phi"] = flex.double([0.0, 45.0, 90.0])
    rt["xyzobs.px.value"] = flex.vec3_double(
        [(0.0, 0.0, 0.0), (0.0, 0.0, 45.0), (0.0, 0.0, 90.0)]
    )
    return rt


@pytest.fixture
def simple_reflection_table():
    """Create a small reflection table"""
    refl = flex.reflection_table()
    refl["intensity"] = flex.double([1.0, 2.0, 3.0])
    refl["variance"] = flex.double([1.0, 2.0, 3.0])
    refl["d"] = flex.double([1.0, 2.0, 3.0])
    refl["miller_index"] = flex.miller_index([(0, 0, 3), (0, 0, 2), (0, 0, 1)])
    refl.set_flags(flex.bool(refl.size(), False), refl.flags.bad_for_scaling)
    return refl


def refl_for_norm():
    """Create 11000 refelctions in 10 groups of 1100 approx equally spaced in
    resolution."""
    intensity_array = flex.double([])
    miller_indices = flex.miller_index([])
    # a set of miller indices with h2 + k2 + l2 = [2,3,4,5,6,8,9,10,11,12],
    # which should split nicely into 10 resolution groups.
    miller_array_list = [
        (1, 1, 0),
        (1, 1, 1),
        (2, 0, 0),
        (2, 1, 0),
        (2, 1, 1),
        (2, 2, 0),
        (2, 2, 1),
        (3, 0, 1),
        (3, 1, 1),
        (2, 2, 2),
    ]
    for i in range(1, 11):
        miller_indices.extend(flex.miller_index(1100, miller_array_list[i - 1]))
        intensity_array.extend(
            flex.double(np.linspace(90, 110, num=1100, endpoint=True))
        )
    rt = flex.reflection_table()
    rt["intensity"] = intensity_array
    rt["variance"] = intensity_array
    rt["miller_index"] = miller_indices
    rt.set_flags(flex.bool(11000, False), rt.flags.bad_for_scaling)
    return rt


def test_quasi_normalisation(simple_reflection_table, test_exp_E2, test_exp_P1):
    """Test the quasi_normalisation function."""
    # Test that for small datasets, all Esq values are set to one.
    refl = quasi_normalisation(simple_reflection_table, test_exp_E2)
    assert list(refl["Esq"]) == [1.0, 1.0, 1.0]

    rt = refl_for_norm()
    new_rt = quasi_normalisation(rt, test_exp_P1)
    for i in range(0, 9):
        assert list(new_rt["Esq"][i * 1100 : (i + 1) * 1100]) == pytest.approx(
            list(np.linspace(0.9, 1.1, num=1100, endpoint=True))
        )
    # FIXME Note, this test should actually be for i in range(0, 10):, however
    # the binner appears to create an extra bin above the highest data,
    # and then the call to interpolate causes the last values to be incorrect.


def test_calc_crystal_frame_vectors_single_axis_gonio(
    test_reflection_table, test_experiment_singleaxisgonio
):
    """Test the namesake function, to check that the vectors are correctly rotated
    into the crystal frame."""
    rt, exp = test_reflection_table, test_experiment_singleaxisgonio
    reflection_table = calc_crystal_frame_vectors(rt, exp)

    # s0c and s1c are normalised. s0c points towards the source.
    # as the crystal rotates about the x axis, the s0 vector moves in the y-z plane towards -y
    expected_s0c = [
        (0.0, 0.0, -1.0),
        (0.0, -1.0 / sqrt(2.0), -1.0 / sqrt(2.0)),
        (0.0, -1.0, 0.0),
    ]
    for v1, v2 in zip(reflection_table["s0c"], expected_s0c):
        assert v1 == pytest.approx(v2)
    # the s1c vector should have fixed x-component, rotating in the y-z plane towards +y
    expected_s1c = [
        (1.0 / sqrt(2.0), 0.0, 1.0 / sqrt(2.0)),
        (1.0 / sqrt(2.0), 0.5, 0.5),
        (1.0 / sqrt(2.0), 1.0 / sqrt(2.0), 0.0),
    ]
    for v1, v2 in zip(reflection_table["s1c"], expected_s1c):
        assert v1 == pytest.approx(v2)

    # now test redefined coordinates so that the lab x-axis is along the
    # z-axis in the crystal frame
    alignment_axis = (1.0, 0.0, 0.0)
    reflection_table["s1c"] = align_axis_along_z(
        alignment_axis, reflection_table["s1c"]
    )
    reflection_table["s0c"] = align_axis_along_z(
        alignment_axis, reflection_table["s0c"]
    )
    expected_s0c_realigned = [
        (1.0, 0.0, 0.0),
        (1.0 / sqrt(2.0), -1.0 / sqrt(2.0), 0.0),
        (0.0, -1.0, 0.0),
    ]
    for v1, v2 in zip(reflection_table["s0c"], expected_s0c_realigned):
        assert v1 == pytest.approx(v2)
    expected_s1c_realigned = [
        (-1.0 / sqrt(2.0), 0.0, 1.0 / sqrt(2.0)),
        (-0.5, 0.5, 1.0 / sqrt(2.0)),
        (0.0, 1.0 / sqrt(2.0), 1.0 / sqrt(2.0)),
    ]
    for v1, v2 in zip(reflection_table["s1c"], expected_s1c_realigned):
        assert v1 == pytest.approx(v2)


def test_calc_crystal_frame_vectors_multi_axis_gonio(test_reflection_table):
    """Test the namesake function, to check that the vectors are correctly rotated
    into the crystal frame."""
    experiments = test_experiments_multiaxisgonio()
    table = generate_reflection_table()

    # for the first scan, the rotation axis is the (1,0,0) direction, like the
    # single axis gonio test case above, so check that first.

    table = calc_crystal_frame_vectors(table, experiments[0])

    # s0c and s1c are normalised. s0c points towards the source.
    # as the crystal rotates about the x axis, the s0 vector moves in the y-z plane towards -y
    expected_s0c = [
        (0.0, 0.0, -1.0),
        (0.0, -1.0 / sqrt(2.0), -1.0 / sqrt(2.0)),
        (0.0, -1.0, 0.0),
    ]
    for v1, v2 in zip(table["s0c"], expected_s0c):
        assert v1 == pytest.approx(v2)
    # the s1c vector should have fixed x-component, rotating in the y-z plane towards +y
    expected_s1c = [
        (1.0 / sqrt(2.0), 0.0, 1.0 / sqrt(2.0)),
        (1.0 / sqrt(2.0), 0.5, 0.5),
        (1.0 / sqrt(2.0), 1.0 / sqrt(2.0), 0.0),
    ]
    for v1, v2 in zip(table["s1c"], expected_s1c):
        assert v1 == pytest.approx(v2)

    # for second scan, the rotation axis is the (1,1,0) direction
    table = generate_reflection_table().select(flex.bool([True, False, True]))
    table = calc_crystal_frame_vectors(table, experiments[1])

    # s0c and s1c are normalised. s0c points towards the source.
    # as the crystal rotates about the (1,1,0) axis, the s0 vector rotates towards (1, -sqrt2, -1)
    # the y-z plane towards -y
    expected_s0c = [
        (0.0, 0.0, -1.0),
        (0.5, -1.0 / sqrt(2.0), -0.5),
    ]
    for v1, v2 in zip(table["s0c"], expected_s0c):
        assert v1 == pytest.approx(v2)
    # the s1c vector should rotate to +y
    expected_s1c = [
        (1.0 / sqrt(2.0), 0.0, 1.0 / sqrt(2.0)),
        (0.0, 1.0, 0.0),
    ]
    for v1, v2 in zip(table["s1c"], expected_s1c):
        assert v1 == pytest.approx(v2)

    # now test redefined coordinates so that the lab x-axis is along the
    # z-axis in the crystal frame
    alignment_axis = (1.0, 0.0, 0.0)
    table["s1c"] = align_axis_along_z(alignment_axis, table["s1c"])
    table["s0c"] = align_axis_along_z(alignment_axis, table["s0c"])
    expected_s0c_realigned = [
        (1.0, 0.0, 0.0),
        (0.5, -1.0 / sqrt(2.0), 0.5),
    ]
    for v1, v2 in zip(table["s0c"], expected_s0c_realigned):
        assert v1 == pytest.approx(v2)
    # the s1c vector should rotate to +y
    expected_s1c_realigned = [
        (-1.0 / sqrt(2.0), 0.0, 1.0 / sqrt(2.0)),
        (0.0, 1.0, 0.0),
    ]
    for v1, v2 in zip(table["s1c"], expected_s1c_realigned):
        assert v1 == pytest.approx(v2)


def test_create_sph_harm_table(test_reflection_table, test_experiment_singleaxisgonio):
    """Simple test for the spherical harmonic table, constructing the table step
    by step, and verifying the values of a few easy-to-calculate entries.
    This also acts as a test for the calc_theta_phi function as well."""

    rt, exp = test_reflection_table, test_experiment_singleaxisgonio
    reflection_table = calc_crystal_frame_vectors(rt, exp)
    reflection_table["s0c"] = align_axis_along_z(
        (1.0, 0.0, 0.0), reflection_table["s0c"]
    )
    reflection_table["s1c"] = align_axis_along_z(
        (1.0, 0.0, 0.0), reflection_table["s1c"]
    )
    theta_phi = calc_theta_phi(reflection_table["s0c"])
    # so s0c vectors realigned in xyz is
    #    (1.0, 0.0, 0.0),
    #    (1.0 / sqrt(2.0), -1.0 / sqrt(2.0), 0.0),
    #    (0.0, -1.0, 0.0),
    # physics conventions, theta from 0 to pi, phi from 0 to 2pi
    expected = [
        (pi / 2.0, 0.0),
        (pi / 2.0, 7.0 * pi / 4.0),
        (pi / 2.0, 3.0 * pi / 2.0),
    ]
    for v1, v2 in zip(theta_phi, expected):
        assert v1 == pytest.approx(v2)
    theta_phi_2 = calc_theta_phi(reflection_table["s1c"])
    expected = [
        (pi / 4.0, pi),
        (pi / 4.0, 3 * pi / 4.0),
        (pi / 4.0, 1.0 * pi / 2.0),
    ]
    for v1, v2 in zip(theta_phi_2, expected):
        assert v1 == pytest.approx(v2)
    sph_h_t = create_sph_harm_table(theta_phi, theta_phi_2, 2)
    Y10 = ((3.0 / (8.0 * pi)) ** 0.5) / 2.0
    Y20 = -1.0 * ((5.0 / (256.0 * pi)) ** 0.5)
    assert sph_h_t[1, 0] == pytest.approx(Y10)
    assert sph_h_t[1, 1] == pytest.approx(Y10)
    assert sph_h_t[1, 2] == pytest.approx(Y10)
    assert sph_h_t[5, 0] == pytest.approx(Y20)
    assert sph_h_t[5, 1] == pytest.approx(Y20)
    assert sph_h_t[5, 2] == pytest.approx(Y20)
    # Now test that you get the same by just calling the function.


def test_calculate_wilson_outliers(wilson_test_reflection_table):
    """Test the set wilson outliers function."""
    reflection_table = set_wilson_outliers(wilson_test_reflection_table)

    assert list(
        reflection_table.get_flags(reflection_table.flags.outlier_in_scaling)
    ) == [True, False, True, False]


def test_calculate_prescaling_correction():
    """Test the helper function that applies the lp, dqe and partiality corr."""
    reflection_table = flex.reflection_table()
    reflection_table["lp"] = flex.double([1.0, 0.9, 0.8])
    reflection_table["qe"] = flex.double([0.6, 0.5, 0.4])

    reflection_table = calculate_prescaling_correction(reflection_table)
    assert list(reflection_table["prescaling_correction"]) == [
        1.0 / 0.6,
        0.9 / 0.5,
        0.8 / 0.4,
    ]

    # Test compatibility for old datasets
    del reflection_table["qe"]
    reflection_table["dqe"] = flex.double([0.6, 0.5, 0.4])
    reflection_table = calculate_prescaling_correction(reflection_table)
    assert list(reflection_table["prescaling_correction"]) == [
        1.0 / 0.6,
        0.9 / 0.5,
        0.8 / 0.4,
    ]


def test_reasons():
    """Test the reasons class, which is basically a dictionary with a nice
    printed output"""
    reasons = Reasons()
    reasons.add_reason("test reason", 100)
    assert reasons.reasons["test reason"] == 100
    print(reasons)
    expected_output = """Reflections passing individual criteria:
criterion: test reason, reflections: 100
"""
    assert reasons.__repr__() == expected_output


def test_calculate_harmonic_tables_from_selections():
    selection = flex.size_t([1, 0, 2, 3, 1])
    coefficients = [flex.double([10, 11, 12, 13]), flex.double([20, 21, 22, 23])]

    arrays, mat = calculate_harmonic_tables_from_selections(
        selection, selection, coefficients
    )
    assert len(arrays) == 2
    assert mat.n_cols == 2
    assert mat.n_rows == 5
    assert list(arrays[0]) == [11, 10, 12, 13, 11]
    assert list(arrays[1]) == [21, 20, 22, 23, 21]
    assert mat[0, 0] == arrays[0][0]
    assert mat[1, 0] == arrays[0][1]
    assert mat[2, 0] == arrays[0][2]
    assert mat[3, 0] == arrays[0][3]
    assert mat[4, 0] == arrays[0][4]
    assert mat[0, 1] == arrays[1][0]
    assert mat[1, 1] == arrays[1][1]
    assert mat[2, 1] == arrays[1][2]
    assert mat[3, 1] == arrays[1][3]
    assert mat[4, 1] == arrays[1][4]


def test_equality_of_two_harmonic_table_methods(dials_data):
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    refl = location / "20_integrated.pickle"
    expt = location / "20_integrated_experiments.json"

    phil_scope = phil.parse(
        """
      include scope dials.command_line.scale.phil_scope
    """,
        process_includes=True,
    )
    parser = ArgumentParser(phil=phil_scope, check_format=False)
    params, _ = parser.parse_args(args=[], quick_parse=True)
    params.model = "physical"
    lmax = 2
    params.physical.lmax = lmax

    reflection_table = flex.reflection_table.from_file(refl)
    experiments = load.experiment_list(expt, check_format=False)
    experiments = create_scaling_model(params, experiments, [reflection_table])

    experiment = experiments[0]
    # New method

    reflection_table["phi"] = (
        reflection_table["xyzobs.px.value"].parts()[2]
        * experiment.scan.get_oscillation()[1]
    )
    reflection_table = calc_crystal_frame_vectors(reflection_table, experiment)
    reflection_table["s1c"] = align_axis_along_z(
        (1.0, 0.0, 0.0), reflection_table["s1c"]
    )
    reflection_table["s0c"] = align_axis_along_z(
        (1.0, 0.0, 0.0), reflection_table["s0c"]
    )
    theta_phi_0 = calc_theta_phi(reflection_table["s0c"])  # array of tuples in radians
    theta_phi_1 = calc_theta_phi(reflection_table["s1c"])
    points_per_degree = 4
    s0_lookup_index = calc_lookup_index(theta_phi_0, points_per_degree)
    s1_lookup_index = calc_lookup_index(theta_phi_1, points_per_degree)
    print(list(s0_lookup_index[0:20]))
    print(list(s1_lookup_index[0:20]))
    coefficients_list = create_sph_harm_lookup_table(lmax, points_per_degree)
    experiment.scaling_model.components["absorption"].data = {
        "s0_lookup": s0_lookup_index,
        "s1_lookup": s1_lookup_index,
    }
    experiment.scaling_model.components[
        "absorption"
    ].coefficients_list = coefficients_list
    assert experiment.scaling_model.components["absorption"]._mode == "memory"
    experiment.scaling_model.components["absorption"].update_reflection_data()
    absorption = experiment.scaling_model.components["absorption"]
    harmonic_values_list = absorption.harmonic_values[0]

    experiment.scaling_model.components["absorption"].parameters = flex.double(
        [0.1, -0.1, 0.05, 0.02, 0.01, -0.05, 0.12, -0.035]
    )
    scales, derivatives = experiment.scaling_model.components[
        "absorption"
    ].calculate_scales_and_derivatives()

    # Old method:

    old_data = {"sph_harm_table": create_sph_harm_table(theta_phi_0, theta_phi_1, lmax)}
    experiment.scaling_model.components["absorption"].data = old_data
    assert experiment.scaling_model.components["absorption"]._mode == "speed"
    experiment.scaling_model.components["absorption"].update_reflection_data()
    old_harmonic_values = absorption.harmonic_values[0]
    for i in range(0, 8):
        print(i)
        assert list(harmonic_values_list[i]) == pytest.approx(
            list(old_harmonic_values.col(i).as_dense_vector()), abs=0.01
        )

    experiment.scaling_model.components["absorption"].parameters = flex.double(
        [0.1, -0.1, 0.05, 0.02, 0.01, -0.05, 0.12, -0.035]
    )
    scales_1, derivatives_1 = experiment.scaling_model.components[
        "absorption"
    ].calculate_scales_and_derivatives()

    assert list(scales_1) == pytest.approx(list(scales), abs=0.001)
    assert list(scales_1) != [1.0] * len(scales_1)


def test_calc_lookup_index():
    pi = 3.141
    theta_phi = flex.vec2_double(
        [(0.001, 0.0), (pi, 2.0 * pi), (0.001, 2.0 * pi), (pi, 0)]
    )
    indices = calc_lookup_index(theta_phi, 1)
    assert list(indices) == [0, 64799, 359, 64440]
    indices = calc_lookup_index(theta_phi, 2)
    assert list(indices) == [0, 259199, 719, 258480]



from __future__ import annotations

import os

import pytest

from dxtbx.serialize import load

from dials.array_family import flex


@pytest.fixture
def multi_sequence_data(dials_regression):
    experiments = load.experiment_list(
        os.path.join(
            dials_regression, "indexing_test_data", "multi_sweep", "experiments.json"
        ),
        check_format=False,
    )
    reflections = flex.reflection_table.from_file(
        os.path.join(
            dials_regression, "indexing_test_data", "multi_sweep", "indexed.pickle"
        )
    )
    return {"reflections": reflections, "experiments": experiments}


@pytest.fixture
def centroid_test_data(dials_data):
    experiments = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json",
        check_format=False,
    )
    reflections = flex.reflection_table.from_file(
        dials_data("centroid_test_data", pathlib=True) / "integrated.pickle"
    )
    return {"reflections": reflections, "experiments": experiments}



from __future__ import annotations

import collections
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

import numpy as np
import scipy.spatial.distance as ssd
from scipy.cluster import hierarchy

from cctbx import crystal, uctbx
from cctbx.sgtbx.lattice_symmetry import metric_subgroups
from cctbx.uctbx.determine_unit_cell import NCDist
from dxtbx.util import format_float_with_standard_uncertainty

if TYPE_CHECKING:
    import matplotlib.axes


logger = logging.getLogger(__name__)


class Cluster:
    def __init__(
        self,
        crystal_symmetries: list[crystal.symmetry],
        lattice_ids: Optional[list[int]] = None,
        name="",
    ):
        assert lattice_ids is None or len(crystal_symmetries) == len(lattice_ids)
        self.crystal_symmetries = crystal_symmetries
        self.lattice_ids = lattice_ids
        self.name = name
        self.unit_cells = np.array(
            [cs.unit_cell().parameters() for cs in crystal_symmetries]
        )
        self.median_cell = np.median(self.unit_cells, axis=0).tolist()
        self.cell_std = np.std(self.unit_cells, axis=0).tolist()
        self.mean_cell = np.mean(self.unit_cells, axis=0).tolist()
        self.pg_composition = collections.Counter(
            cs.space_group().type().lookup_symbol() for cs in crystal_symmetries
        )

    def __len__(self):
        return len(self.crystal_symmetries)

    def __str__(self):
        cell_str = ", ".join(
            format_float_with_standard_uncertainty(v, e, minimum=1e-5)
            for (v, e) in zip(self.median_cell, self.cell_std)
        )
        return "\n".join(
            [
                f"Median cell: {cell_str} ",
            ]
        )


@dataclass
class ClusteringResult:
    clusters: list[Cluster]
    dendrogram: Optional[dict] = None
    linkage_matrix: Optional[np.ndarray] = None

    def __len__(self):
        return len(self.clusters)

    def __str__(self):
        text = [
            "\n{:<16} {:<8} {:<13} {:<13} {:<13} {:<12} {:<12} {:<12}{:<8}".format(
                "Cluster_id",
                "N_xtals",
                "Med_a",
                "Med_b",
                "Med_c",
                "Med_alpha",
                "Med_beta",
                "Med_gamma",
                "Delta(deg)",
            )
        ]
        singletons = []
        for cluster in self.clusters:
            if len(cluster) > 1:
                input_symmetry = crystal.symmetry(
                    unit_cell=uctbx.unit_cell(cluster.median_cell),
                    space_group_symbol="P 1",
                )
                groups = metric_subgroups(
                    input_symmetry, 3.00, enforce_max_delta_for_generated_two_folds=True
                )
                group = groups.result_groups[0]
                uc_params_conv = group["best_subsym"].unit_cell().parameters()
                sorted_pg_comp = sorted(
                    cluster.pg_composition.items(), key=lambda x: -1 * x[1]
                )
                pg_strings = ["{} in {}".format(pg[1], pg[0]) for pg in sorted_pg_comp]
                point_group_string = ", ".join(pg_strings) + "."
                text.append(point_group_string)
                text.append(
                    (
                        "{:<16} {:<8} {:<6.2f}({:<5.2f}) {:<6.2f}({:<5.2f})"
                        " {:<6.2f}({:<5.2f}) {:<6.2f}({:<4.2f}) {:<6.2f}"
                        "({:<4.2f}) {:<6.2f}({:<4.2f})"
                    ).format(
                        cluster.name,
                        len(cluster),
                        cluster.median_cell[0],
                        cluster.cell_std[0],
                        cluster.median_cell[1],
                        cluster.cell_std[1],
                        cluster.median_cell[2],
                        cluster.cell_std[2],
                        cluster.median_cell[3],
                        cluster.cell_std[3],
                        cluster.median_cell[4],
                        cluster.cell_std[4],
                        cluster.median_cell[5],
                        cluster.cell_std[5],
                    )
                )
                text.append(
                    (
                        "{:>24}  {:<6.2f}{:<7} {:<6.2f}{:<7}"
                        " {:<6.2f}{:<7} {:<6.2f}{:<6} {:<6.2f}"
                        "{:<6} {:<6.2f}{:<6}  {:<6.2}"
                    ).format(
                        group["best_subsym"].space_group_info().symbol_and_number(),
                        uc_params_conv[0],
                        "",
                        uc_params_conv[1],
                        "",
                        uc_params_conv[2],
                        "",
                        uc_params_conv[3],
                        "",
                        uc_params_conv[4],
                        "",
                        uc_params_conv[5],
                        "",
                        group["max_angular_difference"],
                    )
                )
            else:
                singletons.append(
                    "".join(
                        [
                            (
                                "{:<14} {:<11.2f} {:<11.2f} {:<11.2f}"
                                "{:<12.1f} {:<12.1f} {:<12.1f}"
                            ).format(
                                list(cluster.pg_composition.keys())[0],
                                cluster.median_cell[0],
                                cluster.median_cell[1],
                                cluster.median_cell[2],
                                cluster.median_cell[3],
                                cluster.median_cell[4],
                                cluster.median_cell[5],
                            ),
                        ]
                    )
                )
        text.append("Standard deviations are in brackets.")
        text.append(
            """Each cluster:
        Input lattice count, with integration Bravais setting space group.
        Cluster median with Niggli cell parameters (std dev in brackets).
        Highest possible metric symmetry and unit cell using LePage (J Appl Cryst 1982, 15:255) method, maximum delta 3deg."""
        )
        n_clusters = len(self.clusters) - len(singletons)
        text.insert(0, f"\n{n_clusters} cluster{'s'[:n_clusters^1]}")
        n_singletons = len(singletons)
        text[:0] = [
            f"{n_singletons} singleton{'s'[:n_singletons^1]}:\n",
            "{:<14} {:<11} {:<11} {:<11}{:<12} {:<12} {:<12}".format(
                "Point group", "a", "b", "c", "alpha", "beta", "gamma"
            ),
            "".join(singletons),
        ]
        return "\n".join(text)


def cluster_unit_cells(
    crystal_symmetries: list[crystal.symmetry],
    lattice_ids: Optional[list[int]] = None,
    threshold: int = 10000,
    ax: Optional["matplotlib.axes.Axes"] = None,
    no_plot: bool = True,
) -> Optional[ClusteringResult]:
    if not lattice_ids:
        lattice_ids = list(range(len(crystal_symmetries)))
    cluster = Cluster(crystal_symmetries, lattice_ids)
    uc = cluster.unit_cells

    a = uc[:, 0] ** 2
    b = uc[:, 1] ** 2
    c = uc[:, 2] ** 2
    d = 2 * uc[:, 1] * uc[:, 2] * np.cos(np.radians(uc[:, 3]))
    e = 2 * uc[:, 0] * uc[:, 2] * np.cos(np.radians(uc[:, 4]))
    f = 2 * uc[:, 0] * uc[:, 1] * np.cos(np.radians(uc[:, 5]))
    g6_cells = np.array([a, b, c, d, e, f]).transpose()

    logger.info(
        "Using Andrews-Bernstein distance from Andrews & Bernstein "
        "J Appl Cryst 47:346 (2014)"
    )
    metric = NCDist
    pair_distances = ssd.pdist(g6_cells, metric=metric)
    if len(pair_distances) > 0:
        logger.info("Distances have been calculated")
        linkage_matrix = hierarchy.linkage(
            pair_distances, method="single", metric=metric
        )
        cluster_ids = hierarchy.fcluster(
            linkage_matrix, threshold, criterion="distance"
        )
        logger.debug("Clusters have been calculated")
    else:
        logger.debug("No distances were calculated. Aborting clustering.")
        return None

    # Create an array of sub-cluster objects from the clustering
    sub_clusters: list[Cluster] = []
    for i_cluster in set(cluster_ids):
        sel = np.where(cluster_ids == i_cluster)[0]
        cluster = Cluster(
            [crystal_symmetries[i] for i in sel],
            [lattice_ids[int(i)] for i in sel],
        )
        sub_clusters.append(cluster)

    # Order clusters by size
    sub_clusters = sorted(sub_clusters, key=len)
    for i, cluster in enumerate(sub_clusters):
        cluster.name = f"cluster_{i + 1}"

    dendrogram = hierarchy.dendrogram(
        linkage_matrix,
        # labels=labels,
        p=200,
        truncate_mode="lastp",  # show only the last p merged clusters
        leaf_font_size=8,
        leaf_rotation=90.0,
        color_threshold=threshold,
        ax=ax,
        no_plot=no_plot,
    )

    return ClusteringResult(
        clusters=sub_clusters, dendrogram=dendrogram, linkage_matrix=linkage_matrix
    )



from __future__ import annotations


def version():
    import os
    import sys

    import dials
    from dials.util.version import dials_version

    print(dials_version())
    print("Python {0.major}.{0.minor}.{0.micro}".format(sys.version_info))
    print(f"Installed in: {os.path.split(dials.__file__)[0]}")


def run(args=None):
    import dials.util

    with dials.util.show_mail_handle_errors():
        version()


if __name__ == "__main__":
    run()



"""
Tests for intensity combination.
"""

from __future__ import annotations

from unittest.mock import Mock

import pytest

from dxtbx.model import Crystal, Experiment

from dials.algorithms.scaling.combine_intensities import (
    MultiDatasetIntensityCombiner,
    SingleDatasetIntensityCombiner,
    combine_intensities,
)
from dials.algorithms.scaling.scaling_utilities import calculate_prescaling_correction
from dials.array_family import flex


@pytest.fixture(scope="module")
def test_exp_P1():
    """Create a mock experiments object."""
    exp = Experiment()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 1.0],
        "space_group_hall_symbol": " P 1",
    }
    crystal = Crystal.from_dict(exp_dict)
    exp.crystal = crystal
    return exp


def generate_simple_table(prf=True):
    """Generate a reflection table for testing intensity combination.
    The numbers are contrived to make sum intensities agree well at high
    intensity but terribly at low and vice versa for profile intensities."""
    reflections = flex.reflection_table()
    reflections["miller_index"] = flex.miller_index(
        [
            (0, 0, 1),
            (0, 0, 1),
            (0, 0, 1),
            (0, 0, 1),
            (0, 0, 1),
            (0, 0, 2),
            (0, 0, 2),
            (0, 0, 2),
            (0, 0, 2),
            (0, 0, 2),
            (0, 0, 3),
            (0, 0, 3),
            (0, 0, 3),
            (0, 0, 3),
            (0, 0, 3),
            (0, 0, 4),
            (0, 0, 4),
            (0, 0, 4),
            (0, 0, 4),
            (0, 0, 4),
            (0, 0, 5),
            (0, 0, 5),
            (0, 0, 5),
            (0, 0, 5),
            (0, 0, 5),
        ]
    )
    reflections["inverse_scale_factor"] = flex.double(25, 1.0)
    # Contrive an example that should give the best cc12 when combined.
    # make sum intensities agree well at high intensity but terribly at low
    # and vice versa for profile intensities.
    # profile less consistent at high intensity here

    # sumless consistent at low intensity here
    reflections["intensity.sum.value"] = flex.double(
        [
            10000.0,
            11000.0,
            9000.0,
            8000.0,
            12000.0,
            500.0,
            5600.0,
            5500.0,
            2000.0,
            6000.0,
            100.0,
            50.0,
            150.0,
            75.0,
            125.0,
            30.0,
            10.0,
            2.0,
            35.0,
            79.0,
            1.0,
            10.0,
            20.0,
            10.0,
            5.0,
        ]
    )
    reflections["intensity.sum.variance"] = flex.double(
        [10000] * 5 + [5000] * 5 + [100] * 5 + [30] * 5 + [10] * 5
    )
    reflections.set_flags(flex.bool(25, False), reflections.flags.outlier_in_scaling)
    reflections.set_flags(flex.bool(25, True), reflections.flags.integrated)
    reflections["lp"] = flex.double(25, 0.5)
    if prf:
        reflections["intensity.prf.value"] = flex.double(
            [
                10000.0,
                16000.0,
                12000.0,
                6000.0,
                9000.0,
                5000.0,
                2000.0,
                1500.0,
                1300.0,
                9000.0,
                100.0,
                80.0,
                120.0,
                90.0,
                100.0,
                30.0,
                40.0,
                50.0,
                30.0,
                30.0,
                10.0,
                12.0,
                9.0,
                8.0,
                10.0,
            ]
        )
        reflections["intensity.prf.variance"] = flex.double(
            [1000] * 5 + [500] * 5 + [10] * 5 + [3] * 5 + [1] * 5
        )
    reflections = calculate_prescaling_correction(reflections)
    return reflections


def test_combine_intensities_prf_sum(test_exp_P1):
    reflections = flex.reflection_table()

    reflections["intensity.sum.value"] = flex.double([100.0, 100.0, 100.0, 100.0])
    reflections["intensity.prf.value"] = flex.double([200.0, 200.0, 200.0, 200.0])
    reflections["intensity.sum.variance"] = flex.double(4, 100)
    reflections["intensity.prf.variance"] = flex.double(4, 200)
    reflections["prescaling_correction"] = flex.double(4, 1.0)
    reflections.set_flags(
        flex.bool([False, False, True, True]), reflections.flags.integrated_prf
    )
    reflections.set_flags(
        flex.bool([True, False, False, True]), reflections.flags.integrated_sum
    )

    intensities, _ = combine_intensities(reflections, Imid=100.0)
    # if prf not successful - set as sum. Only last refl should be combined here.
    assert list(intensities) == [100.0, 100.0, 200.0, 150.0]


def test_combine_intensities(test_exp_P1):
    """Test the combine intensities function for a single dataset"""
    reflections = generate_simple_table()
    scaler = Mock()
    scaler.reflection_table = reflections
    scaler.suitable_refl_for_scaling_sel = flex.bool(reflections.size(), True)
    scaler.outliers = flex.bool(reflections.size(), False)
    scaler.free_set_selection = flex.bool(reflections.size(), False)
    scaler.experiment = test_exp_P1
    scaler.params.reflection_selection.combine.Imid = None

    combiner = SingleDatasetIntensityCombiner(scaler)
    Imid = combiner.max_key
    intensity, variance = combiner.calculate_suitable_combined_intensities()

    # Imid being 1200.0 should be best for this contrived example
    assert Imid == 1200.0

    # Due to nature of crossover, just require 2% tolerance for this example
    assert list(intensity[0:5]) == pytest.approx(
        list(
            reflections["intensity.sum.value"][0:5]
            * reflections["prescaling_correction"][0:5]
        ),
        rel=2e-2,
    )
    assert list(intensity[20:25]) == pytest.approx(
        list(
            reflections["intensity.prf.value"][20:25]
            * reflections["prescaling_correction"][20:25]
        ),
        rel=2e-2,
    )

    assert list(variance[0:5]) == pytest.approx(
        list(
            reflections["intensity.sum.variance"][0:5]
            * flex.pow2(reflections["prescaling_correction"][0:5])
        ),
        rel=2e-2,
    )
    assert list(variance[20:25]) == pytest.approx(
        list(
            reflections["intensity.prf.variance"][20:25]
            * flex.pow2(reflections["prescaling_correction"][20:25])
        ),
        rel=2e-2,
    )

    combiner.max_key = 0  # prf
    intensity, variance = combiner.calculate_suitable_combined_intensities()
    assert list(intensity) == pytest.approx(
        list(reflections["intensity.prf.value"] * reflections["prescaling_correction"]),
        rel=2e-2,
    )
    assert list(variance) == pytest.approx(
        list(
            reflections["intensity.prf.variance"]
            * flex.pow2(reflections["prescaling_correction"])
        ),
        rel=2e-2,
    )

    combiner.max_key = 1  # sum
    intensity, variance = combiner.calculate_suitable_combined_intensities()
    assert list(intensity) == pytest.approx(
        list(reflections["intensity.sum.value"] * reflections["prescaling_correction"]),
        rel=2e-2,
    )
    assert list(variance) == pytest.approx(
        list(
            reflections["intensity.sum.variance"]
            * flex.pow2(reflections["prescaling_correction"])
        ),
        rel=2e-2,
    )


def test_combine_intensities_multi_dataset(test_exp_P1):
    """Test the combine intensities function for multiple datasets"""
    r1 = generate_simple_table()
    r1["partiality"] = flex.double(25, 1.0)
    r2 = generate_simple_table(prf=False)
    scaler1 = Mock()
    scaler1.reflection_table = r1
    scaler1.suitable_refl_for_scaling_sel = flex.bool(r1.size(), True)
    scaler1.outliers = flex.bool(r1.size(), False)
    scaler1.free_set_selection = flex.bool(r1.size(), False)
    scaler1.experiment = test_exp_P1
    scaler1.params.reflection_selection.combine.Imid = None
    scaler2 = Mock()
    scaler2.reflection_table = r2
    scaler2.suitable_refl_for_scaling_sel = flex.bool(r2.size(), True)
    scaler2.outliers = flex.bool(r2.size(), False)
    scaler2.free_set_selection = flex.bool(r2.size(), False)
    scaler2.experiment = test_exp_P1
    scaler2.params.reflection_selection.combine.Imid = None

    multiscaler = Mock()
    multiscaler.active_scalers = [scaler1, scaler2]
    multiscaler.experiment = test_exp_P1
    multiscaler.params.reflection_selection.combine.Imid = None

    combiner = MultiDatasetIntensityCombiner(multiscaler)
    Imid = combiner.max_key

    assert pytest.approx(Imid) == 1200.0



from __future__ import annotations

import math

from cctbx.crystal import symmetry
from libtbx.phil import parse

from dials.array_family import flex
from dials.util import tabulate

"""
Class to find a reasonable cutoff for integration based on work in LABELIT.
Bins the reflection table by resolution, then finds the first bin that goes
under a given I/sigI limit.  Cuts the data at that point.
"""

phil_scope = parse(
    """
significance_filter
  .expert_level=1
{
  enable = False
    .type=bool
    .help=If enabled, the significance filter will, for every experiment, find \
          the highest resolution where the I/sigI remains above a certain point\
          (controlled by isigi_cutoff).
  d_min = None
    .type = float
    .help = High resolution cutoff for binning. If None, use the highest \
            resolution reflection as d_min.
  n_bins = 20
    .type = int
    .help = Number of bins to use when examining resolution falloff
  isigi_cutoff = 1.0
    .type = float
    .help = I/sigI cutoff. Reflections in and past the first bin that falls \
            below this cutoff will not be retained
}
"""
)


class SignificanceFilter:
    def __init__(self, params):
        self.params = params.significance_filter
        self.best_d_min = None

    def __call__(self, experiments, reflections):
        results = flex.reflection_table()
        table_header = ["", "", "", "I", "IsigI", "N >", "RMSD", "Cutoff"]
        table_header2 = [
            "Bin",
            "Resolution Range",
            "Completeness",
            "",
            "",
            "cutoff",
            "(um)",
            "",
        ]

        for exp_id in range(len(experiments)):
            print("*" * 80)
            print("Significance filtering experiment", exp_id)
            table_data = []
            table_data.append(table_header)
            table_data.append(table_header2)
            experiment = experiments[exp_id]

            # Find the bins for this experiment
            crystal = experiment.crystal
            refls = reflections.select(reflections["id"] == exp_id)
            sym = symmetry(
                unit_cell=crystal.get_unit_cell(), space_group=crystal.get_space_group()
            )
            d = crystal.get_unit_cell().d(refls["miller_index"])
            mset = sym.miller_set(indices=refls["miller_index"], anomalous_flag=False)
            binner = mset.setup_binner(
                n_bins=self.params.n_bins, d_min=self.params.d_min or 0
            )
            acceptable_resolution_bins = []

            # Iterate through the bins, examining I/sigI at each bin
            for i in binner.range_used():
                d_max, d_min = binner.bin_d_range(i)
                if d_max < 0:
                    sel = d > d_min
                else:
                    sel = (d <= d_max) & (d > d_min)
                sel &= refls["intensity.sum.value"] > 0
                bin_refls = refls.select(sel)
                n_refls = len(bin_refls)
                avg_i = (
                    flex.mean(bin_refls["intensity.sum.value"]) if n_refls > 0 else 0
                )
                avg_i_sigi = (
                    flex.mean(
                        bin_refls["intensity.sum.value"]
                        / flex.sqrt(bin_refls["intensity.sum.variance"])
                    )
                    if n_refls > 0
                    else 0
                )
                acceptable_resolution_bins.append(
                    avg_i_sigi >= self.params.isigi_cutoff
                )

                bright_refls = bin_refls.select(
                    (
                        bin_refls["intensity.sum.value"]
                        / flex.sqrt(bin_refls["intensity.sum.variance"])
                    )
                    >= self.params.isigi_cutoff
                )
                n_bright = len(bright_refls)

                rmsd_obs = (
                    1000
                    * math.sqrt(
                        (
                            bright_refls["xyzcal.mm"] - bright_refls["xyzobs.mm.value"]
                        ).sum_sq()
                        / n_bright
                    )
                    if n_bright > 0
                    else 0
                )

                table_row = []
                table_row.append("%3d" % i)
                table_row.append(
                    "%-13s"
                    % binner.bin_legend(
                        i_bin=i,
                        show_bin_number=False,
                        show_bin_range=False,
                        show_d_range=True,
                        show_counts=False,
                    )
                )
                table_row.append(
                    "%13s"
                    % binner.bin_legend(
                        i_bin=i,
                        show_bin_number=False,
                        show_bin_range=False,
                        show_d_range=False,
                        show_counts=True,
                    )
                )

                table_row.append(f"{avg_i:.2f}")
                table_row.append(f"{avg_i_sigi:.2f}")
                table_row.append("%3d" % n_bright)
                table_row.append(f"{rmsd_obs:.2f}")
                table_data.append(table_row)

            # Throw out bins that go back above the cutoff after the first non-passing bin is found
            acceptable_resolution_bins = [
                acceptable_resolution_bins[i]
                for i in range(len(acceptable_resolution_bins))
                if False not in acceptable_resolution_bins[: i + 1]
            ]

            for b, row in zip(acceptable_resolution_bins, table_data[2:]):
                if b:
                    row.append("X")
            print(tabulate(table_data, headers="firstrow"))

            # Save the results
            if any(acceptable_resolution_bins):
                best_index = acceptable_resolution_bins.count(True) - 1
                best_row = table_data[best_index + 2]
                d_min = binner.bin_d_range(binner.range_used()[best_index])[1]
                self.best_d_min = d_min
                print("best row:", " ".join(best_row))
                if self.params.enable:
                    results.extend(refls.select(d >= d_min))
            else:
                print("Data didn't pass cutoff")
                self.best_d_min = None
        if self.params.enable:
            return results
        else:
            return reflections



from __future__ import annotations

import json
import os

import procrunner


def test_export_mosflm(dials_regression, tmpdir):
    dials_regression_escaped = json.dumps(dials_regression).strip('"')
    with open(
        os.path.join(dials_regression, "experiment_test_data/experiment_1.json")
    ) as fi:
        with (tmpdir / "experiments.json").open("w") as fo:
            fo.write(fi.read().replace("$DIALS_REGRESSION", dials_regression_escaped))

    result = procrunner.run(
        ["dials.export", "format=mosflm", "experiments.json"], working_directory=tmpdir
    )
    assert not result.returncode and not result.stderr

    assert os.path.exists(tmpdir / "mosflm" / "index.mat")
    with open(tmpdir / "mosflm" / "index.mat") as f:
        lines = f.read()
    assert (
        lines
        == """
 -0.01210200 -0.01954526  0.00309519
 -0.00416605 -0.00080573 -0.02427340
  0.01931593 -0.01241956 -0.00329641
       0.000       0.000       0.000
 -0.52228050 -0.84350975  0.12535704
 -0.17980379 -0.03477015 -0.98308781
  0.83360283 -0.53598726 -0.13350648
     42.2717     42.2720     39.6704     90.0001     89.9993     89.9998
       0.000       0.000       0.000
""".strip(
            "\n"
        )
    )
    assert os.path.exists(tmpdir / "mosflm" / "mosflm.in")
    with open(tmpdir / "mosflm" / "mosflm.in") as f:
        lines = f.read()
    assert (
        lines
        == """
DIRECTORY %s%scentroid_test_data
TEMPLATE centroid_####.cbf
SYMMETRY 89
BEAM 220.002 212.478
DISTANCE 190.1800
MATRIX index.mat
""".strip(
            "\n"
        )
        % (dials_regression, os.path.sep)
    )



from __future__ import annotations

from dials.algorithms.refinement.outlier_detection.outlier_base import CentroidOutlier
from dials.algorithms.statistics.fast_mcd import FastMCD, maha_dist_sq
from dials_refinement_helpers_ext import qchisq


class MCD(CentroidOutlier):
    """Implementation of the CentroidOutlier class using Mahalanobis distance
    calculated using the robust location and scatter estimate from the Minimum
    Covariance Determinant estimate."""

    def __init__(
        self,
        cols=None,
        min_num_obs=20,
        separate_experiments=True,
        separate_panels=True,
        separate_images=False,
        block_width=None,
        nproc=1,
        alpha=0.5,
        max_n_groups=5,
        min_group_size=300,
        n_trials=500,
        k1=2,
        k2=2,
        k3=100,
        threshold_probability=0.975,
    ):

        if cols is None:
            cols = ["x_resid", "y_resid", "phi_resid"]
        CentroidOutlier.__init__(
            self,
            cols=cols,
            min_num_obs=min_num_obs,
            separate_experiments=separate_experiments,
            separate_panels=separate_panels,
            separate_images=separate_images,
            block_width=block_width,
            nproc=nproc,
        )

        # Keep the FastMCD options here
        self._alpha = alpha
        self._max_n_groups = max_n_groups
        self._min_group_size = min_group_size
        self._n_trials = n_trials
        self._k1 = k1
        self._k2 = k2
        self._k3 = k3

        # Calculate Mahalanobis distance threshold
        df = len(cols)
        self._mahasq_cutoff = qchisq(threshold_probability, df)

        return

    def _detect_outliers(self, cols):

        fast_mcd = FastMCD(
            cols,
            alpha=self._alpha,
            max_n_groups=self._max_n_groups,
            min_group_size=self._min_group_size,
            n_trials=self._n_trials,
            k1=self._k1,
            k2=self._k2,
            k3=self._k3,
        )

        # get location and MCD scatter estimate
        T, S = fast_mcd.get_corrected_T_and_S()

        # get squared Mahalanobis distances
        d2s = maha_dist_sq(cols, T, S)

        # compare to the threshold
        outliers = d2s > self._mahasq_cutoff

        return outliers



from __future__ import annotations


class ParameterReporter:
    """
    Keeps a record of all the ModelParameterisations and
    ScanVaryingModelParameterisations present and provides access to their
    Parameters and ScanVaryingParameterSets for reporting purposes.

    It is assumed that the provided model parameterisations will be one of five
    types:

    * Detector parameterisation
    * Beam parameterisation
    * Crystal orientation parameterisation
    * Crystal unit cell parameterisation
    * Goniometer setting parameterisation
    """

    def __init__(
        self,
        detector_parameterisations=None,
        beam_parameterisations=None,
        xl_orientation_parameterisations=None,
        xl_unit_cell_parameterisations=None,
        goniometer_parameterisations=None,
    ):

        if detector_parameterisations is None:
            detector_parameterisations = []
        if beam_parameterisations is None:
            beam_parameterisations = []
        if xl_orientation_parameterisations is None:
            xl_orientation_parameterisations = []
        if xl_unit_cell_parameterisations is None:
            xl_unit_cell_parameterisations = []
        if goniometer_parameterisations is None:
            goniometer_parameterisations = []

        # Keep references to all parameterised models
        self._detector_parameterisations = detector_parameterisations
        self._beam_parameterisations = beam_parameterisations
        self._xl_orientation_parameterisations = xl_orientation_parameterisations
        self._xl_unit_cell_parameterisations = xl_unit_cell_parameterisations
        self._goniometer_parameterisations = goniometer_parameterisations

        self._length = self._len()

    def _len(self):

        length = 0
        for model in self._detector_parameterisations:
            length += model.num_free()
        for model in self._beam_parameterisations:
            length += model.num_free()
        for model in self._xl_orientation_parameterisations:
            length += model.num_free()
        for model in self._xl_unit_cell_parameterisations:
            length += model.num_free()
        for model in self._goniometer_parameterisations:
            length += model.num_free()

        return length

    def __len__(self):
        return self._length

    def _indent(self, string):
        return "\n".join("    " + e for e in str(string).split("\n"))

    def __str__(self):

        s = "Parameter Report:\n"
        if self._detector_parameterisations:
            s += "Detector parameters:\n"
            det_plists = [x.get_params() for x in self._detector_parameterisations]
            params = [x for l in det_plists for x in l]
            for p in params:
                tmp = self._indent(p)
                s += tmp + "\n"

        if self._beam_parameterisations:
            s += "Beam parameters:\n"
            beam_plists = [x.get_params() for x in self._beam_parameterisations]
            params = [x for l in beam_plists for x in l]
            for p in params:
                tmp = self._indent(p)
                s += tmp + "\n"

        if self._xl_orientation_parameterisations:
            s += "Crystal orientation parameters:\n"
            xlo_plists = [
                x.get_params() for x in self._xl_orientation_parameterisations
            ]
            params = [x for l in xlo_plists for x in l]
            for p in params:
                tmp = self._indent(p)
                s += tmp + "\n"

        if self._xl_unit_cell_parameterisations:
            s += "Crystal unit cell parameters:\n"
            xluc_plists = [x.get_params() for x in self._xl_unit_cell_parameterisations]
            params = [x for l in xluc_plists for x in l]
            for p in params:
                tmp = self._indent(p)
                s += tmp + "\n"

        if self._goniometer_parameterisations:
            s += "Goniometer parameters:\n"
            gon_plists = [x.get_params() for x in self._goniometer_parameterisations]
            params = [x for l in gon_plists for x in l]
            for p in params:
                tmp = self._indent(p)
                s += tmp + "\n"

        return s

    def varying_params_vs_image_number(self, image_range):
        """Returns a string which is a table of scan-varying parameter values vs
        image number, if scan-varying parameters are present. Otherwise returns
        None"""

        image_numbers = list(range(image_range[0], image_range[1] + 1))
        columns = [TableColumn("Image", image_numbers)]

        for parameterisation in (
            self._detector_parameterisations
            + self._beam_parameterisations
            + self._xl_orientation_parameterisations
            + self._xl_unit_cell_parameterisations
            + self._goniometer_parameterisations
        ):
            for p in parameterisation.get_params():
                try:
                    vals = [
                        parameterisation.get_smoothed_parameter_value(i, p)
                        for i in image_numbers
                    ]
                    columns.append(TableColumn(p.name_stem, vals))
                except AttributeError:
                    continue

        if len(columns) > 1:
            header = "\t".join([e.title for e in columns])
            text = header + "\n"
            for i in range(len(columns[0])):
                vals = "\t".join(["%.6f" % e.values[i] for e in columns])
                text += vals + "\n"
            return text

        else:
            return None

    def get_params(self, only_free=True):
        """return a concatenated list of parameters from each of the components
        in the global model"""

        global_p_list = []
        for parameterisation in (
            self._detector_parameterisations
            + self._beam_parameterisations
            + self._xl_orientation_parameterisations
            + self._xl_unit_cell_parameterisations
            + self._goniometer_parameterisations
        ):
            global_p_list.extend(parameterisation.get_params(only_free))

        return global_p_list


class TableColumn:
    """Bucket to store data to be used for constructing tables to print."""

    def __init__(self, title, values):

        self._title = title
        self._values = values

    def __len__(self):
        return len(self._values)

    @property
    def title(self):
        return self._title

    @property
    def values(self):
        return self._values



from __future__ import annotations

import logging
from math import pi

from scitbx import sparse
from scitbx.array_family import flex

from dials_refinement_helpers_ext import CalculateCellGradients

logger = logging.getLogger(__name__)

DEG2RAD = pi / 180.0
RAD2DEG = 180.0 / pi


class DerivedParameterTie:
    """Calculate the restraint and gradients for a single derived parameter
    of the model"""

    def __init__(self, target, weight):
        self._target = target
        self._w = weight
        self._dRdp = None

    def residual(self, parameter_value, parameter_gradients):
        """Calculate residual R, cache gradients"""

        d = parameter_value - self._target
        self._dRdp = parameter_gradients

        return d

    def gradient(self):
        """Return dR/dp"""
        return self._dRdp

    def weight(self):
        """Return restraint weight"""
        return self._w


class SingleUnitCellTie:
    """Tie the parameters of a single unit cell model parameterisation to
    target values via least-squares restraints. The restraints will be expressed
    in terms of real space unit cell constants, whilst the underlying parameters
    are encapsulated in the model parameterisation object"""

    def __init__(self, model_parameterisation, target, sigma):
        """model_parameterisation is a CrystalUnitCellParameterisation

        target is a sequence of 6 elements describing the target cell parameters

        sigma is a sequence of 6 elements giving the 'sigma' for each of the
        terms in target, from which weights for the residuals will be calculated.
        Values of zero will remove the restraint for the cell parameter at
        that position"""

        self._xlucp = model_parameterisation
        self._target = target

        assert len(self._target) == 6
        assert len(sigma) == 6
        assert None not in sigma

        # calculate gradients of cell parameters wrt model parameters.
        grads = self._calculate_uc_gradients()

        msg = (
            "Unit cell similarity restraints were requested for both the "
            "{0} and {1} dimensions, however for the crystal in experiment "
            "{2} these are constrained to be equal. Only the strongest "
            "of these restraints will be retained."
        )
        # identify cell dimensions constrained to be equal
        a, b, c, aa, bb, cc = self._xlucp.get_model().get_unit_cell().parameters()
        if abs(a - b) < 1e-10:
            grad_diff = [abs(e1 - e2) for (e1, e2) in zip(grads[0], grads[1])]
            if max(grad_diff) < 1e-10:
                # a and b are equal, therefore keep only the strongest restraint
                logger.debug(msg.format("a", "b", self._xlucp.get_experiment_ids()[0]))
                strong, weak = sorted([sigma[0], sigma[1]])
                if strong == 0.0:
                    strong = weak
                sigma[0] = strong
                sigma[1] = 0.0
        if abs(a - c) < 1e-10:
            grad_diff = [abs(e1 - e2) for (e1, e2) in zip(grads[0], grads[2])]
            if max(grad_diff) < 1e-10:
                # a and c are equal, therefore keep only the strongest restraint
                logger.debug(msg.format("a", "c", self._xlucp.get_experiment_ids()[0]))
                strong, weak = sorted([sigma[0], sigma[2]])
                if strong == 0.0:
                    strong = weak
                sigma[0] = strong
                sigma[2] = 0.0
        if abs(b - c) < 1e-10:
            grad_diff = [abs(e1 - e2) for (e1, e2) in zip(grads[1], grads[2])]
            if max(grad_diff) < 1e-10:
                # b and c are equal, therefore keep only the strongest restraint
                logger.debug(msg.format("b", "c", self._xlucp.get_experiment_ids()[0]))
                strong, weak = sorted([sigma[1], sigma[2]])
                if strong == 0.0:
                    strong = weak
                sigma[1] = strong
                sigma[2] = 0.0

        # A gradient of zero indicates that cell parameter is constrained and thus
        # to be ignored in restraints
        msg = (
            "Unit cell similarity restraints were requested for the {0} "
            "parameter, however for the crystal in experiment {1}, {0} is "
            "constrained. This restraint will be removed."
        )
        _sigma = []
        for i, (sig, grad, pname) in enumerate(
            zip(sigma, grads, ["a", "b", "c", "alpha", "beta", "gamma"])
        ):
            tst = (abs(g) > 1.0e-10 for g in grad)
            if any(tst):
                if sig == 0.0:
                    sig = None
                _sigma.append(sig)
            else:
                logger.debug(msg.format(pname, self._xlucp.get_experiment_ids()[0]))
                _sigma.append(None)

        # For each non-zero sigma create a restraint between the relevant cell
        # parameter and its target value
        self._ties = []
        for t, s in zip(self._target, _sigma):
            if s is not None:
                self._ties.append(DerivedParameterTie(t, 1.0 / s**2))
            else:
                self._ties.append(None)

        # set up empty weights list
        self._weights = []

    def _calculate_uc_gradients(self, sel=[True] * 6):
        """Calculate gradients of the unit cell parameters with respect to
        each of the parameters of the crystal unit cell model parameterisation"""

        B = self._xlucp.get_state()
        dB_dp = flex.mat3_double(self._xlucp.get_ds_dp())

        # Use C++ function for speed
        ccg = CalculateCellGradients(B, dB_dp)

        nparam = len(dB_dp)
        da = list(ccg.da_dp()) if sel[0] else [0.0] * nparam
        db = list(ccg.db_dp()) if sel[1] else [0.0] * nparam
        dc = list(ccg.dc_dp()) if sel[2] else [0.0] * nparam
        daa = list(ccg.daa_dp()) if sel[3] else [0.0] * nparam
        dbb = list(ccg.dbb_dp()) if sel[4] else [0.0] * nparam
        dcc = list(ccg.dcc_dp()) if sel[5] else [0.0] * nparam

        return (da, db, dc, daa, dbb, dcc)

    def residuals(self):
        """Calculate and return the residuals, cache gradients"""

        cell_params = self._xlucp.get_model().get_unit_cell().parameters()

        # gradients of the cell parameters wrt model parameters
        grads = self._calculate_uc_gradients(sel=[t is not None for t in self._ties])

        R = []
        for p, g, t in zip(cell_params, grads, self._ties):
            if t is None:
                continue
            R.append(t.residual(parameter_value=p, parameter_gradients=g))

        return R

    @property
    def num_residuals(self):
        """Get the number of residuals"""
        return len(self._ties) - self._ties.count(None)

    def gradients(self):
        """For each residual, return the gradients dR/dp. Requires residuals to be
        called first"""

        dRdp = []
        for t in self._ties:
            if t is None:
                continue
            dRdp.append(t.gradient())

        return dRdp

    def weights(self):
        """Return the weights for the residuals vector"""

        # the weights do not change so cache them
        if not self._weights:
            self._weights = []
            for t in self._ties:
                if t is None:
                    continue
                self._weights.append(t.weight())

        return self._weights


class MeanUnitCellTie:
    """Tie the parameters of multiple unit cell model parameterisations to
    central values via least-squares restraints. The restraints will be expressed
    in terms of real space unit cell constants, whilst the underlying parameters
    are encapsulated in the model parameterisation objects"""

    def __init__(self, model_parameterisations, sigma):
        """model_parameterisations is a list of CrystalUnitCellParameterisations

        sigma is a sequence of 6 elements giving the 'sigma' for each of the
        unit cell parameters, from which weights for the residuals will be
        calculated. Values of zero in sigma will remove the restraint for the
        cell parameter at that position"""

        self._xlucp = model_parameterisations
        self._nxls = len(model_parameterisations)

        # common factors used in gradient calculations
        self._meangradfac = 1.0 / self._nxls
        self._gradfac = 1.0 - self._meangradfac

        self._weights = []

        # initially want to calculate all gradients
        self._sel = [True] * 6

        # identify any cell dimensions constrained to be equal. If any are and a
        # restraint has been requested for that cell dimension, remove the restraint
        # for all crystals and warn in the log
        msg = (
            "Unit cell similarity restraints were requested for both the "
            "{0} and {1} dimensions, however for the crystal in experiment "
            "{2} these are constrained to be equal. Only the strongest "
            "of these restraints will be retained for all crystals in "
            "the restrained group."
        )
        for ixl, xlucp in enumerate(self._xlucp):
            B = xlucp.get_state()
            dB_dp = flex.mat3_double(xlucp.get_ds_dp())
            ccg = CalculateCellGradients(B, dB_dp)
            grads = [
                ccg.da_dp(),
                ccg.db_dp(),
                ccg.dc_dp(),
                ccg.daa_dp(),
                ccg.dbb_dp(),
                ccg.dcc_dp(),
            ]
            a, b, c, aa, bb, cc = xlucp.get_model().get_unit_cell().parameters()
            if abs(a - b) < 1e-10:
                grad_diff = [abs(e1 - e2) for (e1, e2) in zip(grads[0], grads[1])]
                if max(grad_diff) < 1e-10:
                    # a and b are equal for this crystal, therefore keep only the
                    # strongest requested restraint
                    if sigma[0] > 0.0 and sigma[1] > 0.0:
                        logger.debug(
                            msg.format("a", "b", xlucp.get_experiment_ids()[0])
                        )
                        strong, weak = sorted([sigma[0], sigma[1]])
                        sigma[0] = strong
                        sigma[1] = 0.0
            if abs(a - c) < 1e-10:
                grad_diff = [abs(e1 - e2) for (e1, e2) in zip(grads[0], grads[2])]
                if max(grad_diff) < 1e-10:
                    # a and c are equal for this crystal, therefore keep only the
                    # strongest requested restraint
                    if sigma[0] > 0.0 and sigma[2] > 0.0:
                        logger.debug(
                            msg.format("a", "c", xlucp.get_experiment_ids()[0])
                        )
                        strong, weak = sorted([sigma[0], sigma[2]])
                        sigma[0] = strong
                        sigma[2] = 0.0
            if abs(b - c) < 1e-10:
                grad_diff = [abs(e1 - e2) for (e1, e2) in zip(grads[1], grads[2])]
                if max(grad_diff) < 1e-10:
                    # b and c are equal for this crystal, therefore keep only the
                    # strongest requested restraint
                    if sigma[1] > 0.0 and sigma[2] > 0.0:
                        logger.debug(
                            msg.format("b", "c", xlucp.get_experiment_ids()[0])
                        )
                        strong, weak = sorted([sigma[1], sigma[2]])
                        sigma[1] = strong
                        sigma[2] = 0.0

            # A gradient of zero indicates that cell parameter is constrained and thus
            # to be ignored in restraints
            # _sigma = []
            msg = (
                "Unit cell similarity restraints were requested for the {0} "
                "parameter, however for the crystal in experiment {1}, {0} is "
                "constrained. This restraint will be removed for all crystals in "
                "the restrained group."
            )
            for i, (grad, pname) in enumerate(
                zip(grads, ["a", "b", "c", "alpha", "beta", "gamma"])
            ):
                tst = (abs(g) <= 1.0e-10 for g in grad)
                if all(tst):
                    # this parameter is constrained, so remove any requested restraints
                    # at this position
                    if sigma[i] > 0.0:
                        logger.debug(msg.format(pname, xlucp.get_experiment_ids()[0]))
                        sigma[i] = 0.0

        # set the selection for gradient calculations to the unconstrained parameters
        self._sel = [s > 0.0 for s in sigma]

        self.nrestraints_per_cell = self._sel.count(True)

        # repeat the weights for each unit cell being restrained
        weights = [1.0 / s**2 for s in sigma if s > 0.0]
        weights = [flex.double(self._nxls, w) for w in weights]
        self._weights = weights[0]
        for w in weights[1:]:
            self._weights.extend(w)

    @staticmethod
    def average_fn(vals):
        return flex.mean(vals)

    def residuals(self):
        """Calculate and return the residuals"""

        cells = [
            xlucp.get_model().get_unit_cell().parameters() for xlucp in self._xlucp
        ]
        a, b, c, aa, bb, cc = [flex.double(e) for e in zip(*cells)]
        resid_a = a - self.average_fn(a) if self._sel[0] else None
        resid_b = b - self.average_fn(b) if self._sel[1] else None
        resid_c = c - self.average_fn(c) if self._sel[2] else None
        resid_aa = aa - self.average_fn(aa) if self._sel[3] else None
        resid_bb = bb - self.average_fn(bb) if self._sel[4] else None
        resid_cc = cc - self.average_fn(cc) if self._sel[5] else None

        # collect the residuals for restrained parameters only
        resid = [
            e
            for e in [resid_a, resid_b, resid_c, resid_aa, resid_bb, resid_cc]
            if e is not None
        ]

        # stack the columns
        R = resid[0]
        for r in resid[1:]:
            R.extend(r)

        return R

    @property
    def num_residuals(self):
        """Get the number of residuals"""
        return self.nrestraints_per_cell * self._nxls

    def _construct_grad_block(self, param_grads, i):
        """helper function to construct a block of gradients. The length of
        param_grads is the number of columns of the block. i selects a row of
        interest from the block corresponding to the residual for a particular
        unit cell"""
        mean_grads = param_grads * self._meangradfac
        param_grads *= self._gradfac
        block = sparse.matrix(self._nxls, len(param_grads))
        for j, (g, mg) in enumerate(zip(param_grads, mean_grads)):
            if abs(mg) > 1e-20:  # skip gradients close to zero
                col = flex.double(flex.grid(self._nxls, 1), -1.0 * mg)
                block.assign_block(col, 0, j)
            if abs(g) > 1e-20:  # skip gradient close to zero
                block[i, j] = g
        return block

    def gradients(self):
        """A generator function to return the gradients dR/dp for all the restraints
        referring to a particular crystal's cell parameters. The return value is
        a list of sparse matrices, one for each of the 6 cell parameters being
        restrained. Each sparse matrix has as many columns as the crystal unit
        cell parameterisation has parameters, and as many rows as there are crystals
        being restrained. Gradients of zero are detected and not set in the sparse
        matrices to save memory."""

        for i, xlucp in enumerate(self._xlucp):
            B = xlucp.get_state()
            dB_dp = flex.mat3_double(xlucp.get_ds_dp())
            # Use C++ function for speed
            ccg = CalculateCellGradients(B, dB_dp)
            dRdp = []
            if self._sel[0]:
                dRdp.append(self._construct_grad_block(ccg.da_dp(), i))
            if self._sel[1]:
                dRdp.append(self._construct_grad_block(ccg.db_dp(), i))
            if self._sel[2]:
                dRdp.append(self._construct_grad_block(ccg.dc_dp(), i))
            if self._sel[3]:
                dRdp.append(self._construct_grad_block(ccg.daa_dp(), i))
            if self._sel[4]:
                dRdp.append(self._construct_grad_block(ccg.dbb_dp(), i))
            if self._sel[5]:
                dRdp.append(self._construct_grad_block(ccg.dcc_dp(), i))

            yield dRdp

    def weights(self):
        """Return the weights for the residuals vector"""

        return self._weights


class LowMemoryMeanUnitCellTie(MeanUnitCellTie):
    def _construct_grad_block(self, param_grads, i):
        """helper function to construct a block of gradients. The length of
        param_grads is the number of columns of the block. i selects a row of
        interest from the block corresponding to the residual for a particular
        unit cell"""
        param_grads *= self._gradfac
        block = sparse.matrix(self._nxls, len(param_grads))
        for j, g in enumerate(param_grads):
            if abs(g) > 1e-20:  # skip gradient close to zero
                block[i, j] = g
        return block


class MedianUnitCellTie(MeanUnitCellTie):
    @staticmethod
    def average_fn(vals):
        return flex.median(vals)

    def _construct_grad_block(self, param_grads, i):
        """helper function to construct a block of gradients. The length of
        param_grads is the number of columns of the block. i selects a row of
        interest from the block corresponding to the residual for a particular
        unit cell"""
        # this override removes the product with self._gradfac, which is only
        # relevant for the 'mean' versions of this class.
        # param_grads *= self._gradfac
        block = sparse.matrix(self._nxls, len(param_grads))
        for j, g in enumerate(param_grads):
            if abs(g) > 1e-20:  # skip gradient close to zero
                block[i, j] = g
        return block



# LIBTBX_SET_DISPATCHER_NAME dev.dials.make_sphinx_html

from __future__ import annotations

import argparse
import shutil
import sys
from pathlib import Path

import procrunner

import dials.util


@dials.util.show_mail_handle_errors()
def run(args=None):
    # Navigate to the DIALS root
    dials_dir = Path(dials.__file__).parent.parent.parent
    sphinx_dir = dials_dir / "doc" / "sphinx"
    if not sphinx_dir.is_dir():
        sys.exit(
            "Error: Could not find DIALS sphinx folder at ../../doc/sphinx - are you running from a checkout?"
        )
    tutorial_doc_dir = sphinx_dir / "documentation" / "tutorials"

    parser = argparse.ArgumentParser(
        description="Generate documentation website for DIALS",
        prog="dev.dials.make_sphinx_html",
    )
    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "-s",
        "--strict",
        action="store_true",
        default=True,
        help="Run in strict mode and stop on encountering any errors or warnings (default)",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        dest="strict",
        action="store_false",
        help="Ignore any errors or warnings",
    )
    parser.add_argument(
        "-l",
        "--logs",
        dest="logs",
        type=Path,
        help="Use generated dials output logs from this location.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=sphinx_dir / "build" / "html",
        help="Generate output in this location. Defaults to dials doc/sphinx dir.",
    )
    parser.add_argument(
        "--legacy-version",
        metavar="VERSION",
        help="Add a warning message to every page saying that this documentation "
        "relates to the out-of-date version VERSION",
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        help="Empty the output directory before building the documentation",
    )
    parser.add_argument(
        "--parallel",
        action="store_true",
        help="Build documentation in parallel",
    )
    options = parser.parse_args(args)
    print(options)

    output_dir = Path(options.output).resolve()
    if options.clean:
        print(f"Cleaning out {output_dir}")
        shutil.rmtree(output_dir)
    print(f"Generating documentation into {output_dir}")

    command = ["libtbx.sphinx-build", "-b", "html", ".", output_dir]
    if options.parallel:
        command.extend(["-j", "auto"])
    if options.strict:
        command.append("-W")
    if options.legacy_version:
        command.extend(["-A", "legacy_version=" + options.legacy_version])
    if not options.logs:
        # Try to automatically determine this
        branch = "master"
        if options.legacy_version:
            branch = "dials-" + options.legacy_version
        log_dir = dials_dir / ".." / "dials.github.io" / "tutorial_data" / branch
        if log_dir.is_dir():
            options.logs = Path(log_dir)
            print(f"Using DIALS logs from {options.logs}")

    if options.logs:
        command.extend(["-D", f"dials_logs={options.logs.resolve()}"])
        for report in ("betalactamase", "thaumatin"):
            shutil.copy(
                Path(options.logs) / report / "dials.report.html",
                tutorial_doc_dir / f"{report}-report.html",
            )
    else:
        sys.exit(
            "You must specify the location of the tutorial data output with the '-l' option"
        )

    env = {}
    result = procrunner.run(
        command, environment_override=env, working_directory=sphinx_dir
    )
    if result.returncode:
        sys.exit(f"Sphinx build failed with exit code {result.returncode}")


if __name__ == "__main__":
    run()



from __future__ import annotations

from dials.viewer.bitmap_from_array import wxbmp_from_np_array


class wxbitmap_convert:
    """
    The main duty of this class is to convert
    from
    a 3D flex array or a list of flex arrays
    to
    a list of WxBitmaps
    """

    def __init__(self, data_in_n1, data_in_n2=None):

        if data_in_n1 is None and data_in_n2 is None:

            self.lst_3d_data = None
            self.lst_3d_mask = None

        elif data_in_n1 is not None and data_in_n2 is None:
            self.lst_3d_mask = None
            # No double list given

            if isinstance(data_in_n1, list):
                # is a single list
                self.lst_3d_data = []
                for lst_memb in data_in_n1:
                    self.lst_3d_data.append(lst_memb.as_numpy_array())

            else:
                # Got flex array
                self.lst_3d_data = []
                self.lst_3d_data.append(data_in_n1.as_numpy_array())

        else:
            # Got two arguments
            if isinstance(data_in_n1, list) and isinstance(data_in_n2, list):
                # Got two lists
                if len(data_in_n1) == len(data_in_n2):
                    self.lst_3d_data = []
                    self.lst_3d_mask = []
                    for lst_pos in range(len(data_in_n1)):
                        lst_memb1 = data_in_n1[lst_pos]
                        self.lst_3d_data.append(lst_memb1.as_numpy_array())
                        lst_memb2 = data_in_n2[lst_pos]
                        self.lst_3d_mask.append(lst_memb2.as_numpy_array())

                else:
                    # might be a good idea to raise a runtime error here
                    print("the two lists do NOT have the same size")

            elif not isinstance(data_in_n1, list) and not isinstance(data_in_n2, list):
                # Got two blocks
                self.lst_3d_data = []
                self.lst_3d_mask = []

                img1_np = data_in_n1.as_numpy_array()
                self.lst_3d_data.append(img1_np)

                img2_np = data_in_n2.as_numpy_array()
                self.lst_3d_mask.append(img2_np)

                if (
                    img1_np.shape[0] == 0
                    or img1_np.shape[1] == 0
                    or img1_np.shape[2] == 0
                    or img2_np.shape[0] == 0
                    or img2_np.shape[1] == 0
                    or img2_np.shape[2] == 0
                ):

                    self.lst_3d_data = None
                    self.lst_3d_mask = None

            else:
                # might be a good idea to raise a runtime error here
                print("Got mixture of different type of data")

    def get_wxbitmap_lst(self, show_nums=True, palette="black2white", scale=1.0):
        # returning wxbitmaps needed to be shown
        self.local_bmp = wxbmp_from_np_array(
            self.lst_3d_data, show_nums, palette, self.lst_3d_mask
        )
        return self.scaling(scale)

    def scaling(self, scale=1.0):
        # scaling the list of wxbitmap to be shown
        lst_img = self.local_bmp.bmp_lst_scaled(scale)

        return lst_img



"""Test decomposition of rotation matrices around arbitrary axes"""


from __future__ import annotations

import math
import random

from libtbx.test_utils import approx_equal
from scitbx import matrix

from dials.algorithms.refinement.rotation_decomposition import (
    solve_r3_rotation_for_angles_given_axes,
)


def _test_rotation_matrices(phi1, phi2, phi3):

    # compose rotation matrix
    R1 = matrix.col((1, 0, 0)).axis_and_angle_as_r3_rotation_matrix(phi1, deg=False)
    R2 = matrix.col((0, 1, 0)).axis_and_angle_as_r3_rotation_matrix(phi2, deg=False)
    R3 = matrix.col((0, 0, 1)).axis_and_angle_as_r3_rotation_matrix(phi3, deg=False)
    R = R1 * R2 * R3

    # obtain the solutions
    sol1 = solve_r3_rotation_for_angles_given_axes(
        R, (1, 0, 0), (0, 1, 0), (0, 0, 1), smaller_phi2_solution=False
    )
    sol2 = solve_r3_rotation_for_angles_given_axes(
        R, (1, 0, 0), (0, 1, 0), (0, 0, 1), smaller_phi2_solution=True
    )

    # recompose these into rotation matrices
    tstR1 = matrix.col((1, 0, 0)).axis_and_angle_as_r3_rotation_matrix(
        sol1[0], deg=False
    )
    tstR1 *= matrix.col((0, 1, 0)).axis_and_angle_as_r3_rotation_matrix(
        sol1[1], deg=False
    )
    tstR1 *= matrix.col((0, 0, 1)).axis_and_angle_as_r3_rotation_matrix(
        sol1[2], deg=False
    )

    tstR2 = matrix.col((1, 0, 0)).axis_and_angle_as_r3_rotation_matrix(
        sol2[0], deg=False
    )
    tstR2 *= matrix.col((0, 1, 0)).axis_and_angle_as_r3_rotation_matrix(
        sol2[1], deg=False
    )
    tstR2 *= matrix.col((0, 0, 1)).axis_and_angle_as_r3_rotation_matrix(
        sol2[2], deg=False
    )

    # the two solution sets must reproduce the same rotation matrix
    assert approx_equal(tstR1, tstR2)

    # additionally, this matrix must be equal to the original matrix
    assert approx_equal(tstR1, R)

    # check rotated vectors
    # vec = R * matrix.col((0,0,1))
    # tstvec1 = tstR1 * matrix.col((0,0,1))
    # tstvec2 = tstR2 * matrix.col((0,0,1))
    # print R
    # print tstR1
    # print tstR2
    # print vec
    # print tstvec1
    # print tstvec2


def _test_vs_euler_angles_xyz_angles(phi1, phi2, phi3):

    from scitbx.math import euler_angles_xyz_angles

    # compose rotation matrix
    R1 = matrix.col((1, 0, 0)).axis_and_angle_as_r3_rotation_matrix(phi1, deg=False)
    R2 = matrix.col((0, 1, 0)).axis_and_angle_as_r3_rotation_matrix(phi2, deg=False)
    R3 = matrix.col((0, 0, 1)).axis_and_angle_as_r3_rotation_matrix(phi3, deg=False)
    R = R1 * R2 * R3

    # get two solution sets for the principal axes
    sol1 = solve_r3_rotation_for_angles_given_axes(
        R, (1, 0, 0), (0, 1, 0), (0, 0, 1), smaller_phi2_solution=False, deg=True
    )
    sol2 = solve_r3_rotation_for_angles_given_axes(
        R, (1, 0, 0), (0, 1, 0), (0, 0, 1), smaller_phi2_solution=True, deg=True
    )

    # get the solution set provided by euler_angles_xyz_angles
    tst = euler_angles_xyz_angles(R)

    # one of these must match
    assert approx_equal(sol1, tst, out=None) or approx_equal(sol2, tst, out=None)


def random_vector():
    v = matrix.col((1, 0, 0)).rotate_around_origin(
        matrix.col((0, 1, 0)), random.uniform(0, math.pi)
    )
    return v.rotate_around_origin(
        matrix.col((1, 0, 0)), random.uniform(0, 2.0 * math.pi)
    ).normalize()


def _test_random_axes_and_angles():
    # random axes
    e1, e2, e3 = [random_vector() for i in range(3)]

    # random angles
    phi1 = random.uniform(-math.pi, math.pi)
    phi2 = random.uniform(-math.pi, math.pi)
    phi3 = random.uniform(-math.pi, math.pi)

    # compose rotation matrix
    R1 = e1.axis_and_angle_as_r3_rotation_matrix(phi1, deg=False)
    R2 = e2.axis_and_angle_as_r3_rotation_matrix(phi2, deg=False)
    R3 = e3.axis_and_angle_as_r3_rotation_matrix(phi3, deg=False)
    R = R1 * R2 * R3

    # obtain solution sets
    sol1 = solve_r3_rotation_for_angles_given_axes(R, e1, e2, e3, True)
    sol2 = solve_r3_rotation_for_angles_given_axes(R, e1, e2, e3, False)

    # if either one of the solution sets is not found, indicate that this test
    # was skipped
    if sol1 is None or sol2 is None:
        return "skipped"

    # there are special solutions where rotations 1 and 3 are about the same axis.
    # solve_r3_rotation_for_angles_given_axes provides solutions where phi1 is
    # 0.0 and the rotation is fully expressed by phi3. Skip these degenerate
    # cases too.
    if sol1[0] == 0.0 and sol2[0] == 0.0:
        return "skipped"

    # one of sol1 or sol2 should match
    tst = (phi1, phi2, phi3)
    assert approx_equal(sol1, tst, out=None) or approx_equal(sol2, tst, out=None)


def test():
    n_tests = 100

    # tests using principal axes, and angles in the range +/-30 deg
    for i in range(n_tests):

        phi1 = random.uniform(-math.pi / 6, math.pi / 6)
        phi2 = random.uniform(-math.pi / 6, math.pi / 6)
        phi3 = random.uniform(-math.pi / 6, math.pi / 6)
        _test_rotation_matrices(phi1, phi2, phi3)
        _test_vs_euler_angles_xyz_angles(phi1, phi2, phi3)

    # tests using arbitrary axes and angles, ensuring not too many are skipped
    results = [_test_random_axes_and_angles() for i in range(n_tests)]
    assert results.count("skipped") < 0.2 * n_tests



from __future__ import annotations

import procrunner


def test(run_in_tmp_path):
    from dials.array_family import flex

    table = flex.reflection_table()
    table["hkl"] = flex.miller_index(360)
    table["id"] = flex.int(360)
    table["intensity.sum.value"] = flex.double(360)
    table.as_file("temp1.refl")
    table.as_file("temp2.refl")

    result = procrunner.run(
        [
            "dials.merge_reflection_lists",
            "temp1.refl",
            "temp2.refl",
            "method=update",
        ]
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file("merged.refl")
    assert len(table) == 360

    result = procrunner.run(
        [
            "dials.merge_reflection_lists",
            "temp1.refl",
            "temp2.refl",
            "method=extend",
        ]
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file("merged.refl")
    assert len(table) == 720



from __future__ import annotations


def test_run(dials_data):
    experiments = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    from dxtbx.model.experiment_list import ExperimentListFactory

    experiments = ExperimentListFactory.from_json_file(str(experiments))

    from dials.algorithms.profile_model.modeller import EwaldSphereSampler

    beam = experiments[0].beam
    detector = experiments[0].detector
    goniometer = experiments[0].goniometer
    scan = experiments[0].scan

    sampler = EwaldSphereSampler(beam, detector, goniometer, scan, 1)

    assert sorted(sampler.nearest_n(0)) == sorted([0, 1, 2, 3, 4, 5, 6, 7, 8])

    assert sorted(sampler.nearest_n(1)) == sorted([0, 1, 2, 8, 9, 10])
    assert sorted(sampler.nearest_n(2)) == sorted([0, 2, 3, 1, 11, 12])
    assert sorted(sampler.nearest_n(3)) == sorted([0, 3, 4, 2, 13, 14])
    assert sorted(sampler.nearest_n(4)) == sorted([0, 4, 5, 3, 15, 16])
    assert sorted(sampler.nearest_n(5)) == sorted([0, 5, 6, 4, 17, 18])
    assert sorted(sampler.nearest_n(6)) == sorted([0, 6, 7, 5, 19, 20])
    assert sorted(sampler.nearest_n(7)) == sorted([0, 7, 8, 6, 21, 22])
    assert sorted(sampler.nearest_n(8)) == sorted([0, 8, 1, 7, 23, 24])

    assert sorted(sampler.nearest_n(9)) == sorted([1, 9, 10, 24, 25, 26])
    assert sorted(sampler.nearest_n(10)) == sorted([1, 10, 11, 9, 27, 28])
    assert sorted(sampler.nearest_n(24)) == sorted([8, 24, 9, 23, 55, 56])

    assert sorted(sampler.nearest_n(25)) == sorted([9, 25, 26, 56])
    assert sorted(sampler.nearest_n(26)) == sorted([9, 26, 27, 25])
    assert sorted(sampler.nearest_n(56)) == sorted([24, 56, 25, 55])
